@@ -190,7 +190,20 @@ struct wrapping_mech_info wrapping_tests[] = {
         .wrapping_mech = { CKM_AES_CBC_PAD, aes_iv, sizeof(aes_iv) },
         .wrapping_key_gen_mech = { CKM_AES_KEY_GEN, 0, 0 },
         .sym_keylen = 32,
-    },    {
+    },
+    {
+        .name = "Wrap/Unwrap with AES 128 XTS",
+        .wrapping_mech = { CKM_AES_XTS, aes_iv, sizeof(aes_iv) },
+        .wrapping_key_gen_mech = { CKM_AES_XTS_KEY_GEN, 0, 0 },
+        .sym_keylen = 32,
+    },
+    {
+        .name = "Wrap/Unwrap with AES 256 XTS",
+        .wrapping_mech = { CKM_AES_XTS, aes_iv, sizeof(aes_iv) },
+        .wrapping_key_gen_mech = { CKM_AES_XTS_KEY_GEN, 0, 0 },
+        .sym_keylen = 64,
+    },
+    {
         .name = "Wrap/Unwrap with DES ECB",
         .wrapping_mech = { CKM_DES_ECB, 0, 0 },
         .wrapping_key_gen_mech = { CKM_DES_KEY_GEN, 0, 0 },
@@ -273,6 +286,18 @@ struct wrapped_mech_info wrapped_key_tests[] = {
         .operation_mech = { CKM_AES_ECB, 0, 0 },
         .sym_keylen = 32,
     },
+    {
+        .name = "key type AES-XTS 128",
+        .wrapped_key_gen_mech = { CKM_AES_XTS_KEY_GEN, 0, 0 },
+        .operation_mech = { CKM_AES_XTS, aes_iv, sizeof(aes_iv) },
+        .sym_keylen = 32,
+    },
+    {
+        .name = "key type AES-XTS 256",
+        .wrapped_key_gen_mech = { CKM_AES_XTS_KEY_GEN, 0, 0 },
+        .operation_mech = { CKM_AES_XTS, aes_iv, sizeof(aes_iv) },
+        .sym_keylen = 64,
+    },
     {
         .name = "key type DES3",
         .wrapped_key_gen_mech = { CKM_DES3_KEY_GEN, 0, 0 },
@@ -385,6 +410,7 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
         /* Perform Encrypt/Decrypt operation */
         switch (mech->mechanism) {
         case CKM_AES_ECB:
+        case CKM_AES_XTS:
             input_size = 16;
             encr_key = sym_key1;
             decr_key = sym_key2;
@@ -531,14 +557,14 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
                    mech_to_str(wrap_mech->mechanism));
 
     if (!mech_supported(slot_id1, tsuite->wrapped_key_gen_mech.mechanism)) {
-        testcase_skip("Slot %u doesn't support %s (%u)",
+        testcase_skip("Slot %u doesn't support %s (0x%x)",
                       (unsigned int)slot_id1,
                       mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
                       (unsigned int)tsuite->wrapped_key_gen_mech.mechanism);
         goto testcase_cleanup;
     }
     if (!mech_supported(slot_id2, tsuite->wrapped_key_gen_mech.mechanism)) {
-        testcase_skip("Slot %u doesn't support %s (%u)",
+        testcase_skip("Slot %u doesn't support %s (0x%x)",
                       (unsigned int)slot_id2,
                       mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
                       (unsigned int)tsuite->wrapped_key_gen_mech.mechanism);
@@ -546,14 +572,14 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
     }
 
     if (!mech_supported(slot_id1, tsuite->operation_mech.mechanism)) {
-        testcase_skip("Slot %u doesn't support %s (%u)",
+        testcase_skip("Slot %u doesn't support %s (0x%x)",
                       (unsigned int)slot_id1,
                       mech_to_str(tsuite->operation_mech.mechanism),
                       (unsigned int)tsuite->operation_mech.mechanism);
         goto testcase_cleanup;
     }
     if (!mech_supported(slot_id2, tsuite->operation_mech.mechanism)) {
-        testcase_skip("Slot %u doesn't support %s (%u)",
+        testcase_skip("Slot %u doesn't support %s (0x%x)",
                       (unsigned int)slot_id2,
                       mech_to_str(tsuite->operation_mech.mechanism),
                       (unsigned int)tsuite->operation_mech.mechanism);
@@ -581,30 +607,35 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
 
         if (!keysize_supported(slot_id1, tsuite->wrapped_key_gen_mech.mechanism,
                                tsuite->rsa_modbits)) {
-            testcase_skip("Token in slot %lu cannot be used with "
-                          "modbits.='%ld'", slot_id1, tsuite->rsa_modbits);
+            testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
+                          slot_id1, tsuite->rsa_modbits);
             goto testcase_cleanup;
         }
         if (!keysize_supported(slot_id2, tsuite->wrapped_key_gen_mech.mechanism,
                                tsuite->rsa_modbits)) {
-            testcase_skip("Token in slot %lu cannot be used with "
-                          "modbits.='%ld'", slot_id2, tsuite->rsa_modbits);
+            testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
+                          slot_id2, tsuite->rsa_modbits);
             goto testcase_cleanup;
         }
 
         if (is_ep11_token(slot_id1) || is_ep11_token(slot_id2)) {
             if (!is_valid_ep11_pubexp(tsuite->rsa_publ_exp,
                                       tsuite->rsa_publ_exp_len)) {
-                testcase_skip("EP11 Token in cannot be used with "
-                             "publ_exp.='%s'", s);
+                testcase_skip("EP11 Token in cannot be used with publ_exp.='%s'", s);
                 goto testcase_cleanup;
             }
         }
         if (is_cca_token(slot_id1) || is_cca_token(slot_id2)) {
             if (!is_valid_cca_pubexp(tsuite->rsa_publ_exp,
                                      tsuite->rsa_publ_exp_len)) {
-                testcase_skip("CCA Token in scannot be used with "
-                              "publ_exp.='%s'", s);
+                testcase_skip("CCA Token in scannot be used with publ_exp.='%s'", s);
+                goto testcase_cleanup;
+            }
+        }
+        if (is_soft_token(slot_id1) || is_soft_token(slot_id2)) {
+            if (!is_valid_soft_pubexp(tsuite->rsa_publ_exp,
+                                      tsuite->rsa_publ_exp_len)) {
+                testcase_skip("Soft Token in scannot be used with publ_exp.='%s'", s);
                 goto testcase_cleanup;
             }
         }
@@ -612,8 +643,7 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
             if (!is_valid_tpm_pubexp(tsuite->rsa_publ_exp,
                                      tsuite->rsa_publ_exp_len) ||
                 !is_valid_tpm_modbits(tsuite->rsa_modbits)) {
-                testcase_skip("TPM Token cannot " "be used with "
-                              "publ_exp.='%s'", s);
+                testcase_skip("TPM Token cannot " "be used with publ_exp.='%s'", s);
                 goto testcase_cleanup;
             }
         }
@@ -621,8 +651,7 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
             if (!is_valid_icsf_pubexp(tsuite->rsa_publ_exp,
                                       tsuite->rsa_publ_exp_len) ||
                 tsuite->rsa_modbits < 1024) {
-                testcase_skip("ICSF Token cannot be used with "
-                              "publ_exp='%s'.", s);
+                testcase_skip("ICSF Token cannot be used with publ_exp='%s'.", s);
                 goto testcase_cleanup;
             }
         }
@@ -634,6 +663,13 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         break;
 
     case CKM_AES_KEY_GEN:
+    case CKM_AES_XTS_KEY_GEN:
+        if ((is_ep11_token(slot_id1) || is_ep11_token(slot_id2)) &&
+            tsuite->wrapped_key_gen_mech.mechanism == CKM_AES_XTS_KEY_GEN) {
+            testcase_skip("Skipping as AES XTS is supported only with protected keys");
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
         rc = generate_AESKey(session1, tsuite->sym_keylen, CK_TRUE,
                              &tsuite->wrapped_key_gen_mech, &sym_key);
         break;
@@ -652,7 +688,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
 
     case CKM_EC_KEY_PAIR_GEN:
         rc = generate_EC_KeyPair(session1, tsuite->ec_params,
-                                 tsuite->ec_params_len, &publ_key, &priv_key);
+                                 tsuite->ec_params_len, &publ_key, &priv_key,
+                                 CK_TRUE); // must be extractable for Wrap/Unwrap
         break;
 
     default:
@@ -663,6 +700,16 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
     }
 
     if (rc != CKR_OK) {
+        if (rc == CKR_POLICY_VIOLATION) {
+            testcase_skip("generate to be wrapped key with mech %s (%u) in slot "
+                          "%lu is not allowed by policy",
+                          mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapped_key_gen_mech.mechanism,
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+
         testcase_error("generate to be wrapped key with mech %s (%u) in slot "
                        "%lu failed, rc=%s",
                        mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
@@ -737,6 +784,7 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
     /* Get class and key type from original key */
     switch (tsuite->wrapped_key_gen_mech.mechanism) {
     case CKM_AES_KEY_GEN:
+    case CKM_AES_XTS_KEY_GEN:
     case CKM_GENERIC_SECRET_KEY_GEN:
         unwrap_tmpl_num = 3;
         break;
@@ -807,8 +855,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         }
         if ((rc == CKR_MECHANISM_INVALID || rc == CKR_KEY_TYPE_INCONSISTENT) &&
             is_ep11_token(slot_id2) && wrap_mech->mechanism == CKM_DES3_CBC &&
-            key_type == CKK_EC) {
-            testcase_skip("EP11 does not support unwrap of EC keys with DES3 CBC");
+            (key_type == CKK_EC || key_type == CKK_RSA)) {
+            testcase_skip("EP11 does not support unwrap of EC or RSA keys with DES3 CBC");
             rc = CKR_OK;
             goto testcase_cleanup;
         }
@@ -820,7 +868,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         goto testcase_cleanup;
     }
 
-    if (key_type == CKK_AES || key_type == CKK_GENERIC_SECRET) {
+    if (key_type == CKK_AES || key_type == CKK_AES_XTS ||
+        key_type == CKK_GENERIC_SECRET) {
         /* Check if the unwrapped key has the desired key length */
         rc = funcs->C_GetAttributeValue(session2, unwrapped_key,
                                         getattr_tmpl, 1);
@@ -889,9 +938,9 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     char *s = NULL;
     CK_BYTE modulus[512];
     CK_BYTE publ_exp[16];
-    CK_BYTE key[32];
+    CK_BYTE key[64];
     CK_ULONG key_size = 0;
-    CK_BYTE value[32];
+    CK_BYTE value[64];
     CK_ATTRIBUTE rsa_publ_tmpl[] = {
         {CKA_MODULUS, modulus, sizeof(modulus) },
         {CKA_PUBLIC_EXPONENT, publ_exp, sizeof(publ_exp) },
@@ -903,7 +952,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     testsuite_begin("%s", tsuite->name);
 
     if (!mech_supported(slot_id1, tsuite->wrapping_key_gen_mech.mechanism)) {
-        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (%u)",
+        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (0x%x)",
                        (unsigned int)slot_id1,
                        mech_to_str(tsuite->wrapping_key_gen_mech.mechanism),
                        (unsigned int)tsuite->wrapping_key_gen_mech.mechanism);
@@ -911,7 +960,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     }
 
     if (!mech_supported(slot_id2, tsuite->wrapping_key_gen_mech.mechanism)) {
-        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (%u)",
+        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (0x%x)",
                        (unsigned int)slot_id2,
                        mech_to_str(tsuite->wrapping_key_gen_mech.mechanism),
                        (unsigned int)tsuite->wrapping_key_gen_mech.mechanism);
@@ -919,7 +968,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     }
 
     if (!mech_supported(slot_id1, tsuite->wrapping_mech.mechanism)) {
-        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (%u)",
+        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (0x%x)",
                        (unsigned int)slot_id1,
                        mech_to_str(tsuite->wrapping_mech.mechanism),
                        (unsigned int)tsuite->wrapping_mech.mechanism);
@@ -927,14 +976,14 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     }
     if (!wrap_supported(slot_id1, tsuite->wrapping_mech)) {
         testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support key "
-                       "wrapping with %s (%u)", (unsigned int)slot_id1,
+                       "wrapping with %s (0x%x)", (unsigned int)slot_id1,
                        mech_to_str(tsuite->wrapping_mech.mechanism),
                        (unsigned int)tsuite->wrapping_mech.mechanism);
         goto testcase_cleanup;
     }
 
     if (!mech_supported(slot_id2, tsuite->wrapping_mech.mechanism)) {
-        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (%u)",
+        testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support %s (0x%x)",
                        (unsigned int)slot_id2,
                        mech_to_str(tsuite->wrapping_mech.mechanism),
                        (unsigned int)tsuite->wrapping_mech.mechanism);
@@ -942,7 +991,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     }
     if (!unwrap_supported(slot_id2, tsuite->wrapping_mech)) {
         testsuite_skip(NUM_WRAPPED_KEY_TESTS, "Slot %u doesn't support key "
-                       "wrapping with %s (%u)", (unsigned int)slot_id2,
+                       "wrapping with %s (0x%x)", (unsigned int)slot_id2,
                        mech_to_str(tsuite->wrapping_mech.mechanism),
                        (unsigned int)tsuite->wrapping_mech.mechanism);
         goto testcase_cleanup;
@@ -967,31 +1016,36 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
         if (!keysize_supported(slot_id1,
                                tsuite->wrapping_key_gen_mech.mechanism,
                                tsuite->rsa_modbits)) {
-            testcase_skip("Token in slot %ld cannot be used with "
-                          "modbits.='%ld'", slot_id1, tsuite->rsa_modbits);
+            testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
+                          slot_id1, tsuite->rsa_modbits);
             goto testcase_cleanup;
         }
         if (!keysize_supported(slot_id2,
                                tsuite->wrapping_key_gen_mech.mechanism,
                                tsuite->rsa_modbits)) {
-            testcase_skip("Token in slot %ld cannot be used with "
-                          "modbits.='%ld'", slot_id2, tsuite->rsa_modbits);
+            testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
+                          slot_id2, tsuite->rsa_modbits);
             goto testcase_cleanup;
         }
 
         if (is_ep11_token(slot_id1) || is_ep11_token(slot_id2)) {
             if (!is_valid_ep11_pubexp(tsuite->rsa_publ_exp,
                                       tsuite->rsa_publ_exp_len)) {
-                testcase_skip("EP11 Token in cannot be used with "
-                             "publ_exp.='%s'", s);
+                testcase_skip("EP11 Token in cannot be used with publ_exp.='%s'", s);
                 goto testcase_cleanup;
             }
         }
         if (is_cca_token(slot_id1) || is_cca_token(slot_id2)) {
             if (!is_valid_cca_pubexp(tsuite->rsa_publ_exp,
                                      tsuite->rsa_publ_exp_len)) {
-                testcase_skip("CCA Token in scannot be used with "
-                              "publ_exp.='%s'", s);
+                testcase_skip("CCA Token in scannot be used with publ_exp.='%s'", s);
+                goto testcase_cleanup;
+            }
+        }
+        if (is_soft_token(slot_id1) || is_soft_token(slot_id2)) {
+            if (!is_valid_soft_pubexp(tsuite->rsa_publ_exp,
+                                      tsuite->rsa_publ_exp_len)) {
+                testcase_skip("Soft Token in scannot be used with publ_exp.='%s'", s);
                 goto testcase_cleanup;
             }
         }
@@ -999,8 +1053,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
             if (!is_valid_tpm_pubexp(tsuite->rsa_publ_exp,
                                      tsuite->rsa_publ_exp_len) ||
                 !is_valid_tpm_modbits(tsuite->rsa_modbits)) {
-                testcase_skip("TPM Token cannot " "be used with "
-                              "publ_exp.='%s'", s);
+                testcase_skip("TPM Token cannot " "be used with publ_exp.='%s'", s);
                 goto testcase_cleanup;
             }
         }
@@ -1008,8 +1061,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
             if (!is_valid_icsf_pubexp(tsuite->rsa_publ_exp,
                                       tsuite->rsa_publ_exp_len) ||
                 tsuite->rsa_modbits < 1024) {
-                testcase_skip("ICSF Token cannot be used with "
-                              "publ_exp='%s'.", s);
+                testcase_skip("ICSF Token cannot be used with publ_exp='%s'.", s);
                 goto testcase_cleanup;
             }
         }
@@ -1020,14 +1072,37 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
         break;
 
     case CKM_AES_KEY_GEN:
-        key_size = 32;
+        key_size = tsuite->sym_keylen;
         rc = funcs->C_GenerateRandom(session2, key, key_size);
         if (rc != CKR_OK) {
             testcase_error("C_GenerateRandom(), rc=%s.", p11_get_ckr(rc));
             goto testcase_cleanup;
         }
 
-        rc = create_AESKey(session2, CK_TRUE, key, key_size, &sym_wrap_key2);
+        rc = create_AESKey(session2, CK_TRUE, key, key_size, CKK_AES,
+                           &sym_wrap_key2);
+        if (rc == CKR_POLICY_VIOLATION) {
+            testcase_skip("AES key import is not allowed by policy");
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+        break;
+
+    case CKM_AES_XTS_KEY_GEN:
+        key_size = tsuite->sym_keylen;
+        rc = funcs->C_GenerateRandom(session2, key, key_size);
+        if (rc != CKR_OK) {
+            testcase_error("C_GenerateRandom(), rc=%s.", p11_get_ckr(rc));
+            goto testcase_cleanup;
+        }
+
+        rc = create_AESKey(session2, CK_TRUE, key, key_size, CKK_AES_XTS,
+                           &sym_wrap_key2);
+        if (rc == CKR_POLICY_VIOLATION) {
+            testcase_skip("AES-XTS key import is not allowed by policy");
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
         break;
 
     case CKM_DES3_KEY_GEN:
@@ -1071,6 +1146,16 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     }
 
     if (rc != CKR_OK) {
+        if (rc == CKR_POLICY_VIOLATION) {
+            testcase_skip("generate wrapping key with mech %s (%u) in slot %lu "
+                          "is not allowed by policy",
+                          mech_to_str(tsuite->wrapping_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapping_key_gen_mech.mechanism,
+                          slot_id2);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+
         testcase_error("generate wrapping key with mech %s (%u) in slot %lu "
                        "failed, rc=%s",
                        mech_to_str(tsuite->wrapping_key_gen_mech.mechanism),
@@ -1100,7 +1185,15 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
         sym_tmpl[0].ulValueLen = key_size;
 
         rc = create_AESKey(session1, CK_TRUE, sym_tmpl[0].pValue, sym_tmpl[0].ulValueLen,
-                           &sym_wrap_key1);
+                           CKK_AES, &sym_wrap_key1);
+        break;
+
+    case CKM_AES_XTS_KEY_GEN:
+        memcpy(sym_tmpl[0].pValue, key, key_size);
+        sym_tmpl[0].ulValueLen = key_size;
+
+        rc = create_AESKey(session1, CK_TRUE, sym_tmpl[0].pValue, sym_tmpl[0].ulValueLen,
+                           CKK_AES_XTS, &sym_wrap_key1);
         break;
 
     case CKM_DES3_KEY_GEN:
@@ -1135,6 +1228,13 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     }
 
     if (rc != CKR_OK) {
+        if (rc == CKR_POLICY_VIOLATION) {
+            testcase_skip("import wrapping key in slot %lu is not allowed by policy",
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+
         testcase_error("import wrapping key in slot %lu failed, rc=%s",
                        slot_id1, p11_get_ckr(rc));
         goto testcase_cleanup;
@@ -1142,6 +1242,18 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
 
     /* Wrap/unwrap different keys with this wrapping key */
     for (i = 0; i< NUM_WRAPPED_KEY_TESTS; i++) {
+        /* Some combinations can not work due to size restrictions */
+        if (wrapped_key_tests[i].wrapped_key_gen_mech.mechanism == CKM_AES_XTS_KEY_GEN &&
+            ((tsuite->wrapping_mech.mechanism == CKM_RSA_PKCS &&
+              tsuite->rsa_modbits <= 512) ||
+             (tsuite->wrapping_mech.mechanism == CKM_RSA_PKCS_OAEP &&
+              tsuite->rsa_modbits <= 1024)))
+                continue;
+
+        if (wrapped_key_tests[i].wrapped_key_gen_mech.mechanism == CKM_DES_KEY_GEN &&
+            tsuite->wrapping_mech.mechanism == CKM_AES_XTS)
+            continue;
+
         rc = do_wrap_key_test(&wrapped_key_tests[i], &tsuite->wrapping_mech);
         if (rc != CKR_OK)
             break;
@@ -1185,7 +1297,7 @@ testcase_cleanup:
     return rc;
 }
 
-CK_RV do_tok2tok_tests()
+CK_RV do_tok2tok_tests(void)
 {
     CK_ULONG i;
     CK_RV rc;
@@ -1246,7 +1358,7 @@ int main(int argc, char **argv)
         goto out;
     }
 
-    testcase_setup(0);
+    testcase_setup();
     testcase_begin("Starting...");
 
     // Initialize
@@ -1327,5 +1439,5 @@ finalize:
     }
 out:
     testcase_print_result();
-    return ret;
+    return testcase_return(ret);
 }
