@@ -153,8 +153,10 @@ void CloseMe(STDLL_TokData_t *tokdata, void *node_value,
          * validity, since if we're here, they must already have been valid */
         sltp = &(Anchor->SltList[closeme_arg->slot_id]);
         fcn = sltp->FcnList;
+        BEGIN_HSM_MK_CHANGE_LOCK(sltp, rv)
         rv = fcn->ST_CloseSession(sltp->TokData, s,
                                   closeme_arg->in_fork_initializer);
+        END_HSM_MK_CHANGE_LOCK(sltp, rv)
         if (rv == CKR_OK) {
             decr_sess_counts(closeme_arg->slot_id, s->rw_session);
             bt_node_free(&(Anchor->sess_btree), node_handle, TRUE);
@@ -295,6 +297,68 @@ void decr_sess_counts(CK_SLOT_ID slotID, CK_BBOOL rw_session)
     ProcUnLock();
 }
 
+uint32_t get_tokspec_count(STDLL_TokData_t *tokdata)
+{
+    Slot_Mgr_Shr_t *shm;
+    uint32_t ret;
+
+    shm = Anchor->SharedMemP;
+    if (ProcLock() != CKR_OK)
+        return 0;
+
+    ret = shm->slot_global_tokspec_count[tokdata->slot_id];
+
+    ProcUnLock();
+
+    return ret;
+}
+
+void incr_tokspec_count(STDLL_TokData_t *tokdata)
+{
+    Slot_Mgr_Shr_t *shm;
+#ifdef PKCS64
+    Slot_Mgr_Proc_t_64 *procp;
+#else
+    Slot_Mgr_Proc_t *procp;
+#endif
+
+    // Get the slot mutex
+    shm = Anchor->SharedMemP;
+
+    ProcLock();
+
+    shm->slot_global_tokspec_count[tokdata->slot_id]++;
+
+    procp = &shm->proc_table[Anchor->MgrProcIndex];
+    procp->slot_tokspec_count[tokdata->slot_id]++;
+
+    ProcUnLock();
+}
+
+void decr_tokspec_count(STDLL_TokData_t *tokdata)
+{
+    Slot_Mgr_Shr_t *shm;
+#ifdef PKCS64
+    Slot_Mgr_Proc_t_64 *procp;
+#else
+    Slot_Mgr_Proc_t *procp;
+#endif
+
+    // Get the slot mutex
+    shm = Anchor->SharedMemP;
+
+    ProcLock();
+
+    if (shm->slot_global_tokspec_count[tokdata->slot_id] > 0)
+        shm->slot_global_tokspec_count[tokdata->slot_id]--;
+
+    procp = &shm->proc_table[Anchor->MgrProcIndex];
+    if (procp->slot_tokspec_count[tokdata->slot_id] > 0)
+        procp->slot_tokspec_count[tokdata->slot_id]--;
+
+    ProcUnLock();
+}
+
 // Check if any sessions from other applicaitons exist on this particular
 // token.... This will also validate our own sessions as well.
 // There might be an issue with the fact that a session is created but the
@@ -462,10 +526,10 @@ void DL_UnLoad(API_Slot_t *sltp, CK_SLOT_ID slotID, CK_BBOOL inchildforkinit)
 #endif
 
     if (sltp->TokData) {
-#ifdef ENABLE_LOCKS
         pthread_rwlock_destroy(&sltp->TokData->sess_list_rwlock);
-#endif
         pthread_mutex_destroy(&sltp->TokData->login_mutex);
+        if (sltp->TokData->hsm_mk_change_supported)
+            pthread_rwlock_destroy(&sltp->TokData->hsm_mk_change_rwlock);
         free(sltp->TokData);
         sltp->TokData = NULL;
     }
@@ -607,12 +671,25 @@ int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID, policy_t policy,
     sltp->TokData->real_pid = Anchor->ClientCred.real_pid;
     sltp->TokData->real_uid = Anchor->ClientCred.real_uid;
     sltp->TokData->real_gid = Anchor->ClientCred.real_gid;
+    sltp->TokData->tokspec_counter.get_tokspec_count = get_tokspec_count;
+    sltp->TokData->tokspec_counter.incr_tokspec_count = incr_tokspec_count;
+    sltp->TokData->tokspec_counter.decr_tokspec_count = decr_tokspec_count;
     sltp->TokData->ro_session_count = 0;
     sltp->TokData->global_login_state = CKS_RO_PUBLIC_SESSION;
-#ifdef ENABLE_LOCKS
-    pthread_rwlock_init(&sltp->TokData->sess_list_rwlock, NULL);
-#endif
-    pthread_mutex_init(&sltp->TokData->login_mutex, NULL);
+    sltp->TokData->spinxplfd = -1;
+    sltp->TokData->spinxplfd_count = 0;
+    if (pthread_rwlock_init(&sltp->TokData->sess_list_rwlock, NULL) != 0) {
+        TRACE_ERROR("Initializing session list lock failed.\n");
+        free(sltp->TokData);
+        sltp->TokData = NULL;
+        return FALSE;
+    }
+    if (pthread_mutex_init(&sltp->TokData->login_mutex, NULL) != 0) {
+        TRACE_ERROR("Initializing login mutex failed.\n");
+        free(sltp->TokData);
+        sltp->TokData = NULL;
+        return FALSE;
+    }
     sltp->TokData->policy = policy;
     sltp->TokData->mechtable_funcs = &mechtable_funcs;
     sltp->TokData->statistics = statistics;
@@ -631,6 +708,8 @@ int DL_Load_and_Init(API_Slot_t *sltp, CK_SLOT_ID slotID, policy_t policy,
             DL_Load(sinfp, sltp, dllload);
         }
     } else {
+        free(sltp->TokData);
+        sltp->TokData = NULL;
         return FALSE;
     }
 
