@@ -288,6 +288,16 @@ typedef void (*CSNBKTC_t) (long *return_code,
                            unsigned char *rule_array,
                            unsigned char *key_identifier);
 
+/* Key Token Change 2 */
+typedef void (*CSNBKTC2_t) (long *return_code,
+                            long *reason_code,
+                            long *exit_data_length,
+                            unsigned char *exit_data,
+                            long *rule_array_count,
+                            unsigned char *rule_array,
+                            long *key_identifier_length,
+                            unsigned char *key_identifier);
+
 /* Key Translate */
 typedef void (*CSNBKTR_t) (long *return_code,
                            long *reason_code,
@@ -305,6 +315,18 @@ typedef void (*CSNBRNG_t) (long *return_code,
                            unsigned char *exit_data,
                            unsigned char *form, unsigned char *random_number);
 
+/* Random Number Generate Long */
+typedef void (*CSNBRNGL_t) (long *return_code,
+                           long *reason_code,
+                           long *exit_data_length,
+                           unsigned char *exit_data,
+                           long *rule_array_count,
+                           unsigned char *rule_array,
+                           long *seed_length,
+                           unsigned char *seed,
+                           long *random_number_length,
+                           unsigned char *random_number);
+
 typedef void (*CSNBSAE_t) (long *return_code,
                            long *reason_code,
                            long *exit_data_length,
@@ -1264,3 +1286,37 @@ typedef void (*CSNBHMV_t) (long *return_code,
                            long *chaining_vector_length,
                            unsigned char *chaining_vector,
                            long *MAC_length, unsigned char *MAC_text);
+
+/* Cipher Text Translate 2 */
+typedef void (*CSNBCTT2_t)(long *return_code,
+                           long *reason_code,
+                           long *exit_data_length,
+                           unsigned char *exit_data,
+                           long *rule_array_count,
+                           unsigned char *rule_array,
+                           long *key_identifier_in_length,
+                           unsigned char *key_identifier_in,
+                           long *init_vector_in_length,
+                           unsigned char *init_vector_in,
+                           long *cipher_text_in_length,
+                           unsigned char *cipher_text_in,
+                           long *chaining_vector_length,
+                           unsigned char *chaining_vector,
+                           long *key_identifier_out_length,
+                           unsigned char *key_identifier_out,
+                           long *init_vector_out_length,
+                           unsigned char *init_vector_out,
+                           long *cipher_text_out_length,
+                           unsigned char *cipher_text_out,
+                           long *reserved1_length,
+                           unsigned long *reserved1,
+                           long *reserved2_length,
+                           unsigned char *reserved2);
+
+/* Cryptographic Facility Version */
+typedef void (*CSUACFV_t)(long *return_code,
+                          long *reason_code,
+                          long *exit_data_length,
+                          unsigned char *exit_data,
+                          long *version_data_length,
+                          unsigned char *version_data);
