@@ -212,6 +212,15 @@ typedef struct _OBJECT {
 
     // policy support (set via store_object_strength_f pointer)
     struct objstrength strength;
+
+    /* Allow to attach external data to an object */
+    void *ex_data;
+    size_t ex_data_len;
+    pthread_rwlock_t ex_data_rwlock;
+    void (*ex_data_free)(struct _OBJECT *obj, void *ex_data,
+                         size_t ex_data_len);
+    CK_RV (*ex_data_reload)(struct _OBJECT *obj, void *ex_data,
+                            size_t ex_data_len);
 } OBJECT;
 
 
@@ -344,12 +353,18 @@ CK_RV ock_generic_filter_mechanism_list(STDLL_TokData_t *tokdata,
 /* mech_list.c */
 CK_RV ock_generic_get_mechanism_list(STDLL_TokData_t * tokdata,
                                      CK_MECHANISM_TYPE_PTR pMechanismList,
-                                     CK_ULONG_PTR pulCount);
+                                     CK_ULONG_PTR pulCount,
+                                     CK_BBOOL (*filter_mechanism)
+                                               (STDLL_TokData_t *tokdata,
+                                               CK_MECHANISM_TYPE mechanism));
 
 /* mech_list.c */
 CK_RV ock_generic_get_mechanism_info(STDLL_TokData_t * tokdata,
                                      CK_MECHANISM_TYPE type,
-                                     CK_MECHANISM_INFO_PTR pInfo);
+                                     CK_MECHANISM_INFO_PTR pInfo,
+                                     CK_BBOOL (*filter_mechanism)
+                                               (STDLL_TokData_t *tokdata,
+                                               CK_MECHANISM_TYPE mechanism));
 
 typedef struct _TOK_OBJ_ENTRY {
     CK_BBOOL deleted;
@@ -368,12 +383,19 @@ struct _LW_SHM_TYPE {
     TOK_OBJ_ENTRY priv_tok_objs[MAX_TOK_OBJS];
 };
 
+struct tokspec_counter {
+    uint32_t (*get_tokspec_count)(STDLL_TokData_t *tokdata);
+    void (*incr_tokspec_count)(STDLL_TokData_t *tokdata);
+    void (*decr_tokspec_count)(STDLL_TokData_t *tokdata);
+};
+
 struct _STDLL_TokData_t {
     CK_SLOT_INFO slot_info;
     CK_SLOT_ID slot_id;
     pid_t real_pid; /* pid of client process in pkcsslotd namespace */
     uid_t real_uid; /* uid of client process in pkcsslotd namespace */
     gid_t real_gid; /* gid of client process in pkcsslotd namespace */
+    struct tokspec_counter tokspec_counter;
     int spinxplfd;              // token specific lock
     unsigned int spinxplfd_count; // counter for recursive file lock
     pthread_mutex_t spinxplfd_mutex; // token specific pthread lock
@@ -393,9 +415,7 @@ struct _STDLL_TokData_t {
     unsigned char user_wrap_key[32];
     pthread_mutex_t login_mutex;
     struct btree sess_btree;
-#ifdef ENABLE_LOCKS
     pthread_rwlock_t sess_list_rwlock;
-#endif
     struct btree object_map_btree;
     struct btree sess_obj_btree;
     struct btree publ_token_obj_btree;
@@ -406,6 +426,8 @@ struct _STDLL_TokData_t {
     const struct mechtable_funcs *mechtable_funcs;
     struct statistics *statistics;
     struct tokstore_strength store_strength;
+    CK_BBOOL hsm_mk_change_supported;
+    pthread_rwlock_t hsm_mk_change_rwlock;
 };
 
 #endif
