@@ -857,74 +857,6 @@ done:
     return rc;
 }
 
-CK_RV pkcs_get_keytype(CK_ATTRIBUTE *attrs, CK_ULONG attrs_len,
-                       CK_MECHANISM_PTR mech, CK_ULONG *type, CK_ULONG *class)
-{
-    CK_RV rc;
-
-    *type = 0;
-    *class = 0;
-
-    rc = get_ulong_attribute_by_type(attrs, attrs_len, CKA_CLASS, class);
-    if (rc == CKR_ATTRIBUTE_VALUE_INVALID) {
-        TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
-        return CKR_ATTRIBUTE_VALUE_INVALID;
-    }
-
-    rc = get_ulong_attribute_by_type(attrs, attrs_len, CKA_KEY_TYPE, type);
-    if (rc == CKR_ATTRIBUTE_VALUE_INVALID) {
-        TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
-        return CKR_ATTRIBUTE_VALUE_INVALID;
-    }
-
-    if (rc == CKR_OK)
-        return CKR_OK;
-
-    /* no CKA_KEY_TYPE found, derive from mech */
-    switch (mech->mechanism) {
-    case CKM_DES_KEY_GEN:
-        *type = CKK_DES;
-        break;
-    case CKM_DES2_KEY_GEN:
-        *type = CKK_DES2;
-        break;
-    case CKM_DES3_KEY_GEN:
-        *type = CKK_DES3;
-        break;
-    case CKM_AES_KEY_GEN:
-        *type = CKK_AES;
-        break;
-    case CKM_AES_XTS_KEY_GEN:
-        *type = CKK_AES_XTS;
-        break;
-    case CKM_GENERIC_SECRET_KEY_GEN:
-        *type = CKK_GENERIC_SECRET;
-        break;
-    case CKM_RSA_PKCS_KEY_PAIR_GEN:
-        *type = CKK_RSA;
-        break;
-    case CKM_EC_KEY_PAIR_GEN:
-        *type = CKK_EC;
-        break;
-    case CKM_DSA_KEY_PAIR_GEN:
-        *type = CKK_DSA;
-        break;
-    case CKM_DH_PKCS_KEY_PAIR_GEN:
-        *type = CKK_DH;
-        break;
-    case CKM_IBM_DILITHIUM:
-        *type = CKK_IBM_PQC_DILITHIUM;
-        break;
-    case CKM_IBM_KYBER:
-        *type = CKK_IBM_PQC_KYBER;
-        break;
-    default:
-        return CKR_MECHANISM_INVALID;
-    }
-
-    return CKR_OK;
-}
-
 /**
  * From PKCS#11 v2.40: PKCS #3 Diffie-Hellman key derivation
  *
@@ -1397,11 +1329,15 @@ CK_RV ec_point_from_priv_key(CK_BYTE *parms, CK_ULONG parms_len,
     if (nid == -1)
         return CKR_CURVE_NOT_SUPPORTED;
 
-    bn_d = BN_bin2bn(d, d_len, NULL);
+    bn_d = BN_secure_new();
     if (bn_d == NULL) {
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
+    if (BN_bin2bn(d, d_len, bn_d) == NULL) {
+        rc = CKR_FUNCTION_FAILED;
+        goto done;
+    }
 
     group = EC_GROUP_new_by_curve_name(nid);
     if (group == NULL) {
@@ -1594,3 +1530,73 @@ check_encoded:
 
     return CKR_OK;
 }
+
+/*
+ * Like ec_point_from_public_data(), but returns the EC point in uncompressed
+ * form. If the public data is in compressed form, it uncompresses it.
+ * If it is in hybrid form, convert it into uncompressed form.
+ */
+int ec_point_uncompressed_from_public_data(const CK_BYTE *data,
+                                           CK_ULONG data_len,
+                                           CK_ULONG prime_len,
+                                           CK_BYTE *curve_oid,
+                                           CK_ULONG curve_oid_len,
+                                           CK_BBOOL allow_raw,
+                                           CK_BBOOL *allocated,
+                                           CK_BYTE **ec_point,
+                                           CK_ULONG *ec_point_len)
+{
+    CK_RV rc;
+    CK_BYTE form;
+    CK_ULONG ec_point2_len;
+    CK_BYTE *ec_point2 = NULL;
+
+    rc = ec_point_from_public_data(data, data_len, prime_len, allow_raw,
+                                   allocated, ec_point, ec_point_len);
+    if (rc != CKR_OK)
+        return rc;
+
+    form  = (*ec_point)[0] & ~0x01;
+    switch (form) {
+    case POINT_CONVERSION_COMPRESSED:
+    case POINT_CONVERSION_HYBRID:
+        ec_point2_len = 1 + 2 * prime_len;
+        ec_point2 = malloc(ec_point2_len);
+        if (ec_point2 == NULL) {
+            TRACE_ERROR("Malloc failed\n");
+            rc = CKR_HOST_MEMORY;
+            break;
+        }
+
+        rc = ec_uncompress_public_key(curve_oid, curve_oid_len,
+                                      *ec_point, *ec_point_len, prime_len,
+                                      ec_point2, &ec_point2_len);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Failed to uncompress\n");
+            break;
+        }
+
+        if (*allocated)
+            free(*ec_point);
+        *ec_point = ec_point2;
+        *ec_point_len = ec_point2_len;
+        *allocated = TRUE;
+        ec_point2 = NULL;
+        break;
+
+    case POINT_CONVERSION_UNCOMPRESSED:
+    default:
+        break;
+    }
+
+    if (rc != CKR_OK && *allocated) {
+        free(*ec_point);
+        *ec_point = NULL;
+        *ec_point_len = 0;
+        *allocated = FALSE;
+    }
+    if (ec_point2 != NULL)
+        free(ec_point2);
+
+    return rc;
+}
