@@ -26,8 +26,38 @@
 #include "h_extern.h"
 #include "tok_spec_struct.h"
 #include "trace.h"
+#include "constant_time.h"
 
 #include <openssl/crypto.h>
+#include <openssl/rsa.h>
+
+CK_BBOOL is_rsa_mechanism(CK_MECHANISM_TYPE mech)
+{
+    switch (mech) {
+    case CKM_RSA_9796:
+    case CKM_RSA_PKCS:
+    case CKM_MD2_RSA_PKCS:
+    case CKM_MD5_RSA_PKCS:
+    case CKM_SHA1_RSA_PKCS:
+    case CKM_SHA224_RSA_PKCS:
+    case CKM_SHA256_RSA_PKCS:
+    case CKM_SHA384_RSA_PKCS:
+    case CKM_SHA512_RSA_PKCS:
+    case CKM_RSA_PKCS_PSS:
+    case CKM_SHA1_RSA_PKCS_PSS:
+    case CKM_SHA224_RSA_PKCS_PSS:
+    case CKM_SHA256_RSA_PKCS_PSS:
+    case CKM_SHA384_RSA_PKCS_PSS:
+    case CKM_SHA512_RSA_PKCS_PSS:
+    case CKM_RSA_PKCS_OAEP:
+    case CKM_RSA_X_509:
+    case CKM_RSA_X9_31:
+    case CKM_SHA1_RSA_X9_31:
+        return TRUE;
+    }
+
+    return FALSE;
+}
 
 CK_RV rsa_get_key_info(OBJECT *key_obj, CK_ULONG *mod_bytes,
                        CK_OBJECT_CLASS *keyclass)
@@ -170,10 +200,10 @@ CK_RV rsa_format_block(STDLL_TokData_t *tokdata,
  * 1.5.
  */
 
-CK_RV rsa_parse_block(CK_BYTE *in_data,
-                      CK_ULONG in_data_len,
-                      CK_BYTE *out_data,
-                      CK_ULONG *out_data_len, CK_ULONG type)
+static CK_RV rsa_parse_block_type_1(CK_BYTE *in_data,
+                                    CK_ULONG in_data_len,
+                                    CK_BYTE *out_data,
+                                    CK_ULONG *out_data_len)
 {
     CK_ULONG i;
     CK_RV rc = CKR_OK;
@@ -200,93 +230,45 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
     /*
      * Check the block type.
      */
-    if (in_data[1] != (CK_BYTE) type) {
+    if (in_data[1] != (CK_BYTE) PKCS_BT_1) {
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
         return CKR_ENCRYPTED_DATA_INVALID;
     }
 
     /*
      * The block type shall be a single octet indicating the structure of the
-     * encryption block. It shall have value 00, 01, or 02. For a private-key
-     * operation, the block type shall be 00 or 01. For a public-key
-     * operation, it shall be 02.
+     * encryption block. It shall have value 01. For a private-key
+     * operation, the block type shall be 01.
      *
-     * For block type 00, the octets shall have value 00; for block type 01,
-     * they shall have value FF; and for block type 02, they shall be
-     * pseudorandomly generated and nonzero.
+     * For block type 01, they shall have value FF.
      *
-     * For block type 00, the data must begin with a nonzero octet or have
-     * known length so that the encryption block can be parsed unambiguously.
-     * For block types 01 and 02, the encryption block can be parsed
+     * For block type 01, the encryption block can be parsed
      * unambiguously since the padding string contains no octets with value 00
      * and the padding string is separated from the data by an octet with
      * value 00.
      */
-    switch (type) {
-    /*
-     * For block type 00, the octets shall have value 00.
-     * EB = 00 || 00 || 00 * i || D
-     * Where D must begin with a nonzero octet.
-     */
-    case 0:
-        for (i = 2; i <= (in_data_len - 2); i++) {
-            if (in_data[i] != (CK_BYTE) 0) {
-                found = 1;
-                break;
-            }
-        }
-        if (!found)
-            i++;
-        break;
-    /*
-     * For block type 01, they shall have value FF.
-     * EB = 00 || 01 || FF * i || 00 || D
-     */
-    case 1:
-        for (i = 2; i <= (in_data_len - 2); i++) {
-            if (in_data[i] != (CK_BYTE) 0xff) {
-                if (in_data[i] == (CK_BYTE) 0) {
-                    i++;
-                    found = 1;
-                    break;
-                }
-
-                TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
-                return CKR_ENCRYPTED_DATA_INVALID;
-            }
-        }
-        if (!found)
-            i++;
-        break;
-    /*
-     * For block type 02, they shall be pseudorandomly generated and
-     * nonzero.
-     * EB = 00 || 02 || ?? * i || 00 || D
-     * Where ?? is nonzero.
-     */
-    case 2:
-        for (i = 2; i <= (in_data_len - 2); i++) {
+    for (i = 2; i <= (in_data_len - 2); i++) {
+        if (in_data[i] != (CK_BYTE) 0xff) {
             if (in_data[i] == (CK_BYTE) 0) {
                 i++;
                 found = 1;
                 break;
             }
+
+            TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
+            return CKR_ENCRYPTED_DATA_INVALID;
         }
-        if (!found)
-            i++;
-        break;
-    default:
-        TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
-        return CKR_ENCRYPTED_DATA_INVALID;
     }
+    if (!found)
+        i++;
 
     /*
-     * For block types 01 and 02, the padding string is at least eight octets
+     * For block type 01, the padding string is at least eight octets
      * long, which is a security condition for public-key operations that
      * prevents an attacker from recoving data by trying all possible
      * encryption blocks.
      */
-    if ((type == 1 || type == 2) && ((i - 3) < 8)) {
+    if ((i - 3) < 8) {
         TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
         return CKR_ENCRYPTED_DATA_INVALID;
     }
@@ -307,6 +289,185 @@ CK_RV rsa_parse_block(CK_BYTE *in_data,
     return rc;
 }
 
+#define MAX_LEN_GEN_TRIES 128
+
+static CK_RV rsa_parse_block_type_2(CK_BYTE *in_data,
+                                    CK_ULONG in_data_len,
+                                    CK_BYTE *out_data,
+                                    CK_ULONG *out_data_len,
+                                    CK_BYTE *kdk, CK_ULONG kdklen)
+{
+    unsigned int good = 0, found_zero_byte, equals0;
+    size_t zero_index = 0, msg_index;
+    unsigned char *synthetic = NULL;
+    int synthetic_length;
+    uint16_t len_candidate;
+    unsigned char candidate_lengths[MAX_LEN_GEN_TRIES * sizeof(len_candidate)];
+    uint16_t len_mask;
+    uint16_t max_sep_offset;
+    int synth_msg_index = 0;
+    size_t i, j;
+    CK_RV rc;
+
+    if (kdk == NULL || kdklen == 0) {
+        TRACE_DEVEL("%s\n", ock_err(ERR_ARGUMENTS_BAD));
+        return CKR_ARGUMENTS_BAD;
+    }
+
+    /*
+     * The implementation of this function is copied from OpenSSL's function
+     * ossl_rsa_padding_check_PKCS1_type_2() in crypto/rsa/rsa_pk1.c
+     * and is slightly modified to fit to the OpenCryptoki environment.
+     *
+     * The OpenSSL code is licensed under the Apache License 2.0.
+     * You can obtain a copy in the file LICENSE in the OpenSSL source
+     * distribution or at https://www.openssl.org/source/license.html
+     *
+     * Changes include:
+     * - Different variable, function and define names.
+     * - Usage of TRACE_ERROR to report errors and issue debug messages.
+     * - Different return codes.
+     */
+
+    /*
+     * The format is
+     * 00 || BT || PS || 00 || D
+     * BT - block type
+     * PS - padding string, at least 8 bytes of random non-zero data for BT = 2
+     * D  - data.
+     */
+
+    /*
+     * PKCS#1 v1.5 decryption. See "PKCS #1 v2.2: RSA Cryptography Standard",
+     * section 7.2.2.
+     */
+    if (in_data_len < RSA_PKCS1_PADDING_SIZE) {
+        TRACE_DEVEL("%s\n", ock_err(ERR_FUNCTION_FAILED));
+        return CKR_FUNCTION_FAILED;
+    }
+
+    /* Generate a random message to return in case the padding checks fail. */
+    synthetic = calloc(1, in_data_len);
+    if (synthetic == NULL) {
+        TRACE_ERROR("Failed to allocate synthetic buffer");
+        return CKR_HOST_MEMORY;
+    }
+
+    rc = openssl_specific_rsa_prf(synthetic, in_data_len, "message", 7,
+                                  kdk, kdklen, in_data_len * 8);
+    if (rc != CKR_OK)
+        goto out;
+
+    /* decide how long the random message should be */
+    rc = openssl_specific_rsa_prf(candidate_lengths,
+                                  sizeof(candidate_lengths),
+                                  "length", 6, kdk, kdklen,
+                                  MAX_LEN_GEN_TRIES *
+                                              sizeof(len_candidate) * 8);
+    if (rc != CKR_OK)
+        goto out;
+
+    /*
+     * max message size is the size of the modulus size minus 2 bytes for
+     * version and padding type and a minimum of 8 bytes padding
+     */
+    len_mask = max_sep_offset = in_data_len - 2 - 8;
+    /*
+     * we want a mask so let's propagate the high bit to all positions less
+     * significant than it
+     */
+    len_mask |= len_mask >> 1;
+    len_mask |= len_mask >> 2;
+    len_mask |= len_mask >> 4;
+    len_mask |= len_mask >> 8;
+
+    synthetic_length = 0;
+    for (i = 0; i < MAX_LEN_GEN_TRIES * (int)sizeof(len_candidate);
+                                            i += sizeof(len_candidate)) {
+        len_candidate = (candidate_lengths[i] << 8) |
+                                                candidate_lengths[i + 1];
+        len_candidate &= len_mask;
+
+        synthetic_length = constant_time_select_int(
+                        constant_time_lt(len_candidate, max_sep_offset),
+                                         len_candidate, synthetic_length);
+    }
+
+    synth_msg_index = in_data_len - synthetic_length;
+
+    good = constant_time_is_zero(in_data[0]);
+    good &= constant_time_eq(in_data[1], 2);
+
+    /* scan over padding data */
+    found_zero_byte = 0;
+    for (i = 2; i < in_data_len; i++) {
+        equals0 = constant_time_is_zero(in_data[i]);
+
+        zero_index = constant_time_select_int(~found_zero_byte & equals0,
+                                              i, zero_index);
+        found_zero_byte |= equals0;
+    }
+
+    /*
+     * PS must be at least 8 bytes long, and it starts two bytes into |in_data|.
+     * If we never found a 0-byte, then |zero_index| is 0 and the check
+     * also fails.
+     */
+    good &= constant_time_ge(zero_index, 2 + 8);
+
+    /*
+     * Skip the zero byte. This is incorrect if we never found a zero-byte
+     * but in this case we also do not copy the message out.
+     */
+    msg_index = zero_index + 1;
+
+    /*
+     * old code returned an error in case the decrypted message wouldn't fit
+     * into the |out_data|, since that would leak information, return the
+     * synthetic message instead
+     */
+    good &= constant_time_ge(*out_data_len, in_data_len - msg_index);
+
+    msg_index = constant_time_select_int(good, msg_index, synth_msg_index);
+
+    /*
+     * since at this point the |msg_index| does not provide the signal
+     * indicating if the padding check failed or not, we don't have to worry
+     * about leaking the length of returned message, we still need to ensure
+     * that we read contents of both buffers so that cache accesses don't leak
+     * the value of |good|
+     */
+    for (i = msg_index, j = 0; i < in_data_len && j < *out_data_len;
+                                                                i++, j++)
+        out_data[j] = constant_time_select_8(good, in_data[i], synthetic[i]);
+
+    *out_data_len = j;
+
+out:
+    if (synthetic != NULL)
+        free(synthetic);
+
+    return rc;
+}
+
+CK_RV rsa_parse_block(CK_BYTE *in_data,
+                      CK_ULONG in_data_len,
+                      CK_BYTE *out_data,
+                      CK_ULONG *out_data_len, CK_ULONG type,
+                      CK_BYTE *kdk, CK_ULONG kdklen)
+{
+    switch (type) {
+    case PKCS_BT_1:
+        return rsa_parse_block_type_1(in_data, in_data_len,
+                                        out_data, out_data_len);
+    case PKCS_BT_2:
+        return rsa_parse_block_type_2(in_data, in_data_len,
+                                       out_data, out_data_len, kdk, kdklen);
+    }
+
+    return CKR_ARGUMENTS_BAD;
+}
+
 //
 //
 CK_RV rsa_pkcs_encrypt(STDLL_TokData_t *tokdata,
@@ -454,14 +615,8 @@ CK_RV rsa_pkcs_decrypt(STDLL_TokData_t *tokdata,
     rc = token_specific.t_rsa_decrypt(tokdata, in_data, in_data_len, out_data,
                                       out_data_len, key_obj);
 
-    if (rc != CKR_OK) {
-        if (rc == CKR_DATA_LEN_RANGE) {
-            TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
-            rc = CKR_ENCRYPTED_DATA_LEN_RANGE;
-            goto done;
-        }
-        TRACE_DEVEL("Token Specific rsa decrypt failed.\n");
-    }
+    rc = constant_time_select_int(constant_time_eq(rc, CKR_DATA_LEN_RANGE),
+                                  CKR_ENCRYPTED_DATA_INVALID, rc);
 
 done:
     object_put(tokdata, key_obj, TRUE);
@@ -2377,7 +2532,7 @@ CK_RV ckm_rsa_key_pair_gen(STDLL_TokData_t *tokdata,
     return rc;
 }
 
-CK_RV mgf1(STDLL_TokData_t *tokdata, CK_BYTE *seed, CK_ULONG seedlen,
+CK_RV mgf1(STDLL_TokData_t *tokdata, const CK_BYTE *seed, CK_ULONG seedlen,
            CK_BYTE *mask, CK_ULONG maskLen, CK_RSA_PKCS_MGF_TYPE mgf)
 {
 
@@ -2535,96 +2690,136 @@ CK_RV decode_eme_oaep(STDLL_TokData_t *tokdata, CK_BYTE *emData,
                       CK_ULONG *out_data_len, CK_RSA_PKCS_MGF_TYPE mgf,
                       CK_BYTE *hash, CK_ULONG hlen)
 {
-    int error = 0;;
-    CK_RV rc = CKR_OK;
-    CK_ULONG dbMask_len, ps_len, i;
-    CK_BYTE *maskedSeed, *maskedDB, *dbMask, *seedMask;
+    size_t i, dblen = 0, mlen = -1, one_index = 0, msg_index;
+    unsigned int ok = 0, found_one_byte, mask;
+    const unsigned char *maskedseed, *maskeddb;
+    unsigned char *db = NULL;
+    unsigned char seed[EVP_MAX_MD_SIZE];
 
-    UNUSED(emLen);
+    /*
+     * The implementation of this function is copied from OpenSSL's function
+     * RSA_padding_check_PKCS1_OAEP_mgf1() in crypto/rsa/rsa_oaep.c
+     * and is slightly modified to fit to the OpenCryptoki environment.
+     *
+     * The OpenSSL code is licensed under the Apache License 2.0.
+     * You can obtain a copy in the file LICENSE in the OpenSSL source
+     * distribution or at https://www.openssl.org/source/license.html
+     *
+     * Changes include:
+     * - Different variable and define names.
+     * - Usage of TRACE_ERROR to report errors and issue debug messages.
+     * - Different return codes.
+     * - No need for copying the input to an allocated 'em' buffer. The caller
+     *   guarantees that the size of the input is already the size of the
+     *   modulus.
+     */
 
-    if (!emData || !out_data) {
-        TRACE_ERROR("%s received bad argument(s)\n", __func__);
-        return CKR_FUNCTION_FAILED;
+    /*
+     * |emLen| is guaranteed by the caller to be the modulus size.
+     * |emLen| >= 2 * |hlen| + 2 must hold for the modulus
+     * irrespective of the ciphertext, see PKCS #1 v2.2, section 7.1.2.
+     * This does not leak any side-channel information.
+     */
+    if (emLen < 2 * hlen + 2) {
+        TRACE_ERROR("%s\n", ock_err(ERR_ARGUMENTS_BAD));
+        return CKR_ARGUMENTS_BAD;
     }
 
-    /* allocate memory now for later use */
-    dbMask_len = *out_data_len - hlen - 1;
-    dbMask = malloc(sizeof(CK_BYTE) * dbMask_len);
-    seedMask = malloc(sizeof(CK_BYTE) * hlen);
-    if ((seedMask == NULL) || (dbMask == NULL)) {
+    dblen = emLen - hlen - 1;
+    db = calloc(1, dblen);
+    if (db == NULL) {
         TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
-        rc = CKR_HOST_MEMORY;
-        goto done;
+        return CKR_HOST_MEMORY;
     }
 
-    /* pkcs1v2.2, section 7.1.2, Step 3b:
-     * Separate the encoded message EM and process the decrypted message.
-     *
-     * To mitigate fault and timing attacks, just flag errors and
-     * keep going.
+    /*
+     * The first byte must be zero, however we must not leak if this is
+     * true. See James H. Manger, "A Chosen Ciphertext  Attack on RSA
+     * Optimal Asymmetric Encryption Padding (OAEP) [...]", CRYPTO 2001).
      */
-    maskedSeed = emData + 1;
-    maskedDB = emData + hlen + 1;
+    ok = constant_time_is_zero(emData[0]);
 
-    /* pkcs1v2.2, section 7.1.2, Step 3c:
-     * Compute seedMask using MGF1.
-     */
-    if (mgf1(tokdata, maskedDB, dbMask_len, seedMask, hlen, mgf))
-        error++;
+    maskedseed = emData + 1;
+    maskeddb = emData + 1 + hlen;
 
-    /* pkcs1v2.2, section 7.1.2, Step 3d:
-     * Compute seed using MGF1.
-     */
-    for (i = 0; i < hlen; i++)
-        seedMask[i] ^= maskedSeed[i];
-
-    /* pkcs1v2.2, section 7.1.2, Step 3e:
-     * Compute dbMask using MGF1.
-     */
-    if (mgf1(tokdata, seedMask, hlen, dbMask, dbMask_len, mgf))
-        error++;
-
-    /* pkcs1v2.2, section 7.1.2, Step 3f:
-     * Compute db using MGF1.
-     */
-    for (i = 0; i < dbMask_len; i++)
-        dbMask[i] ^= maskedDB[i];
-
-    /* pkcs1v2.2, section 7.1.2, Step 3g:
-     * DB = lHash’ || PS || 0x01 || M .
-     *
-     * If there is no octet with hexadecimal value 0x01 to separate
-     * PS from M, if lHash does not equal lHash’, output “decryption
-     * error” and stop.
-     */
-    if (memcmp(dbMask, hash, hlen))
-        error++;
-
-    ps_len = hlen;
-    while ((ps_len < dbMask_len) && (dbMask[ps_len] == 0x00))
-        ps_len++;
-
-    if ((ps_len >= dbMask_len) ||
-        (ps_len < dbMask_len && dbMask[ps_len] != 0x01) ||
-        emData[0])
-        error++;
-
-    if (error) {
-        rc = CKR_FUNCTION_FAILED;
+    if (mgf1(tokdata, maskeddb, dblen, seed, hlen, mgf) != CKR_OK)
         goto done;
-    } else {
-        ps_len++;
-        *out_data_len = dbMask_len - ps_len;
-        memcpy(out_data, dbMask + ps_len, dbMask_len - ps_len);
+
+    for (i = 0; i < hlen; i++)
+        seed[i] ^= maskedseed[i];
+
+    if (mgf1(tokdata, seed, hlen, db, dblen, mgf) != CKR_OK)
+        goto done;
+
+    for (i = 0; i < dblen; i++)
+        db[i] ^= maskeddb[i];
+
+    ok &= constant_time_is_zero(CRYPTO_memcmp(db, hash, hlen));
+
+    found_one_byte = 0;
+    for (i = hlen; i < dblen; i++) {
+        /*
+         * Padding consists of a number of 0-bytes, followed by a 1.
+         */
+        unsigned int equals1 = constant_time_eq(db[i], 1);
+        unsigned int equals0 = constant_time_is_zero(db[i]);
+        one_index = constant_time_select_int(~found_one_byte & equals1,
+                                             i, one_index);
+        found_one_byte |= equals1;
+        ok &= (found_one_byte | equals0);
+    }
+
+    ok &= found_one_byte;
+
+    /*
+     * At this point |good| is zero unless the plaintext was valid,
+     * so plaintext-awareness ensures timing side-channels are no longer a
+     * concern.
+     */
+    msg_index = one_index + 1;
+    mlen = dblen - msg_index;
+
+    /*
+     * For good measure, do this check in constant time as well.
+     */
+    ok &= constant_time_ge(*out_data_len, mlen);
+
+    /*
+     * Move the result in-place by |dblen| - |hlen| - 1 - |mlen| bytes
+     * to the left.
+     * Then if |good| move |mlen| bytes from |db| + |hlen| + 1 to |out|.
+     * Otherwise leave |out| unchanged.
+     * Copy the memory back in a way that does not reveal the size of
+     * the data being copied via a timing side channel. This requires copying
+     * parts of the buffer multiple times based on the bits set in the real
+     * length. Clear bits do a non-copy with identical access pattern.
+     * The loop below has overall complexity of O(N*log(N)).
+     */
+    *out_data_len = constant_time_select_int(constant_time_lt(dblen - hlen - 1,
+                                                              *out_data_len),
+                                             dblen - hlen - 1, *out_data_len);
+    for (msg_index = 1; msg_index < dblen - hlen - 1; msg_index <<= 1) {
+        mask = ~constant_time_eq(msg_index & (dblen - hlen - 1 - mlen),
+                                 0);
+        for (i = hlen + 1; i < dblen - msg_index; i++)
+            db[i] = constant_time_select_8(mask, db[i + msg_index], db[i]);
+    }
+    for (i = 0; i < *out_data_len; i++) {
+        mask = ok & constant_time_lt(i, mlen);
+        out_data[i] = constant_time_select_8(mask, db[i + hlen + 1],
+                                             out_data[i]);
     }
 
 done:
-    if (seedMask)
-        free(seedMask);
-    if (dbMask)
-        free(dbMask);
+    OPENSSL_cleanse(seed, sizeof(seed));
+    if (db) {
+        OPENSSL_cleanse(db, dblen);
+        free(db);
+    }
 
-    return rc;
+    *out_data_len = constant_time_select_int(ok, mlen, 0);
+
+    return constant_time_select_int(ok, CKR_OK, CKR_ENCRYPTED_DATA_INVALID);
 }
 
 CK_RV emsa_pss_encode(STDLL_TokData_t *tokdata,
