@@ -82,8 +82,8 @@ struct token_specific_struct {
     CK_RV(*t_set_pin) (STDLL_TokData_t *, SESSION *, CK_CHAR_PTR, CK_ULONG,
                        CK_CHAR_PTR, CK_ULONG);
 
-    CK_RV(*t_des_key_gen) (STDLL_TokData_t *, CK_BYTE **, CK_ULONG *, CK_ULONG,
-                           CK_BBOOL *);
+    CK_RV(*t_des_key_gen) (STDLL_TokData_t *, TEMPLATE *, CK_BYTE **,
+                           CK_ULONG *, CK_ULONG, CK_BBOOL *);
     CK_RV(*t_des_ecb) (STDLL_TokData_t *, CK_BYTE *, CK_ULONG,
                        CK_BYTE *, CK_ULONG *, OBJECT *, CK_BYTE);
     CK_RV(*t_des_cbc) (STDLL_TokData_t *, CK_BYTE *, CK_ULONG,
@@ -209,16 +209,16 @@ struct token_specific_struct {
     CK_RV(*t_generic_secret_key_gen) (STDLL_TokData_t *, TEMPLATE *);
 
     // Token Specific AES functions
-    CK_RV(*t_aes_key_gen) (STDLL_TokData_t *, CK_BYTE **, CK_ULONG *, CK_ULONG,
-                           CK_BBOOL *);
+    CK_RV(*t_aes_key_gen) (STDLL_TokData_t *, TEMPLATE *, CK_BYTE **,
+                           CK_ULONG *, CK_ULONG, CK_BBOOL *);
 
-    CK_RV(*t_aes_xts_key_gen) (STDLL_TokData_t *, CK_BYTE **, CK_ULONG *, CK_ULONG,
-                               CK_BBOOL *);
+    CK_RV(*t_aes_xts_key_gen) (STDLL_TokData_t *, TEMPLATE *, CK_BYTE **,
+                               CK_ULONG *, CK_ULONG, CK_BBOOL *);
 
-    CK_RV(*t_aes_ecb) (STDLL_TokData_t *tokdata, CK_BYTE *, CK_ULONG,
+    CK_RV(*t_aes_ecb) (STDLL_TokData_t *tokdata, SESSION *, CK_BYTE *, CK_ULONG,
                        CK_BYTE *, CK_ULONG *, OBJECT *, CK_BYTE);
 
-    CK_RV(*t_aes_cbc) (STDLL_TokData_t *tokdata, CK_BYTE *, CK_ULONG,
+    CK_RV(*t_aes_cbc) (STDLL_TokData_t *tokdata, SESSION *, CK_BYTE *, CK_ULONG,
                        CK_BYTE *, CK_ULONG *, OBJECT *, CK_BYTE *, CK_BYTE);
 
     CK_RV(*t_aes_ctr) (STDLL_TokData_t *tokdata, CK_BYTE *, CK_ULONG,
@@ -252,7 +252,7 @@ struct token_specific_struct {
     CK_RV(*t_aes_cmac) (STDLL_TokData_t *, CK_BYTE *, CK_ULONG, OBJECT *,
                         CK_BYTE *, CK_BBOOL, CK_BBOOL, CK_VOID_PTR *);
 
-    CK_RV(*t_aes_xts) (STDLL_TokData_t *tokdata, CK_BYTE *, CK_ULONG,
+    CK_RV(*t_aes_xts) (STDLL_TokData_t *tokdata, SESSION *, CK_BYTE *, CK_ULONG,
                        CK_BYTE *, CK_ULONG *, OBJECT *, CK_BYTE *, CK_BBOOL,
                        CK_BBOOL, CK_BBOOL, CK_BYTE*);
 
@@ -290,6 +290,9 @@ struct token_specific_struct {
     CK_RV(*t_handle_event) (STDLL_TokData_t *tokdata, unsigned int event_type,
                             unsigned int event_flags, const char *payload,
                             unsigned int payload_len);
+
+    CK_RV (*t_check_obj_access) (STDLL_TokData_t *tokdata, OBJECT *obj,
+                                 CK_BBOOL create);
 };
 
 typedef struct token_specific_struct token_spec_t;
