@@ -540,6 +540,8 @@ CK_RV token_specific_aes_xts_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl,
                                      CK_BYTE **key, CK_ULONG *len,
                                      CK_ULONG keysize, CK_BBOOL *is_opaque)
 {
+    CK_RV rc;
+
     UNUSED(tmpl);
 
     *key = malloc(keysize);
@@ -548,7 +550,13 @@ CK_RV token_specific_aes_xts_key_gen(STDLL_TokData_t *tokdata, TEMPLATE *tmpl,
     *len = keysize;
     *is_opaque = FALSE;
 
-    return rng_generate(tokdata, *key, keysize);
+    do {
+        rc = rng_generate(tokdata, *key, keysize);
+        if (rc != CKR_OK)
+            return rc;
+    } while (memcmp(*key, (*key) + keysize / 2, keysize / 2) == 0);
+
+    return CKR_OK;
 }
 
 CK_RV token_specific_aes_ecb(STDLL_TokData_t *tokdata, SESSION *sess,
@@ -666,13 +674,15 @@ CK_RV token_specific_aes_cmac(STDLL_TokData_t *tokdata, CK_BYTE *message,
                                      first, last, ctx);
 }
 
-CK_RV token_specific_aes_xts(STDLL_TokData_t *tokdata,
+CK_RV token_specific_aes_xts(STDLL_TokData_t *tokdata, SESSION *sess,
                              CK_BYTE *in_data, CK_ULONG in_data_len,
                              CK_BYTE *out_data, CK_ULONG *out_data_len,
                              OBJECT *key_obj, CK_BYTE *tweak,
                              CK_BOOL encrypt, CK_BBOOL initial, CK_BBOOL final,
                              CK_BYTE* iv)
 {
+    UNUSED(sess);
+
     return openssl_specific_aes_xts(tokdata, in_data, in_data_len,
                                     out_data, out_data_len, key_obj,
                                     tweak, encrypt, initial, final, iv);
@@ -702,7 +712,7 @@ CK_RV token_specific_dh_pkcs_derive(STDLL_TokData_t *tokdata,
 
     //  Create and Init the BIGNUM structures.
     bn_y = BN_new();
-    bn_x = BN_new();
+    bn_x = BN_secure_new();
     bn_p = BN_new();
     bn_z = BN_new();
 
@@ -710,7 +720,7 @@ CK_RV token_specific_dh_pkcs_derive(STDLL_TokData_t *tokdata,
         if (bn_y)
             BN_free(bn_y);
         if (bn_x)
-            BN_free(bn_x);
+            BN_clear_free(bn_x);
         if (bn_p)
             BN_free(bn_p);
         if (bn_z)
@@ -721,20 +731,34 @@ CK_RV token_specific_dh_pkcs_derive(STDLL_TokData_t *tokdata,
     // Initialize context
     ctx = BN_CTX_new();
     if (ctx == NULL) {
+        BN_free(bn_z);
+        BN_free(bn_y);
+        BN_clear_free(bn_x);
+        BN_free(bn_p);
+
         TRACE_ERROR("%s\n", ock_err(ERR_FUNCTION_FAILED));
         return CKR_FUNCTION_FAILED;
     }
-    // Add data into these new BN structures
 
-    BN_bin2bn((unsigned char *) y, y_len, bn_y);
-    BN_bin2bn((unsigned char *) x, x_len, bn_x);
-    BN_bin2bn((unsigned char *) p, p_len, bn_p);
+    // Add data into these new BN structures
+    if (BN_bin2bn((unsigned char *) y, y_len, bn_y) == NULL ||
+        BN_bin2bn((unsigned char *) x, x_len, bn_x) == NULL ||
+        BN_bin2bn((unsigned char *) p, p_len, bn_p) == NULL) {
+        BN_free(bn_z);
+        BN_free(bn_y);
+        BN_clear_free(bn_x);
+        BN_free(bn_p);
+        BN_CTX_free(ctx);
+
+        TRACE_ERROR("%s\n", ock_err(ERR_FUNCTION_FAILED));
+        return CKR_FUNCTION_FAILED;
+    }
 
     rc = BN_mod_exp(bn_z, bn_y, bn_x, bn_p, ctx);
     if (rc == 0) {
         BN_free(bn_z);
         BN_free(bn_y);
-        BN_free(bn_x);
+        BN_clear_free(bn_x);
         BN_free(bn_p);
         BN_CTX_free(ctx);
 
@@ -747,7 +771,7 @@ CK_RV token_specific_dh_pkcs_derive(STDLL_TokData_t *tokdata,
 
     BN_free(bn_z);
     BN_free(bn_y);
-    BN_free(bn_x);
+    BN_clear_free(bn_x);
     BN_free(bn_p);
     BN_CTX_free(ctx);
 
@@ -1064,14 +1088,14 @@ CK_RV token_specific_get_mechanism_list(STDLL_TokData_t *tokdata,
                                         CK_MECHANISM_TYPE_PTR pMechanismList,
                                         CK_ULONG_PTR pulCount)
 {
-    return ock_generic_get_mechanism_list(tokdata, pMechanismList, pulCount);
+    return ock_generic_get_mechanism_list(tokdata, pMechanismList, pulCount, NULL);
 }
 
 CK_RV token_specific_get_mechanism_info(STDLL_TokData_t *tokdata,
                                         CK_MECHANISM_TYPE type,
                                         CK_MECHANISM_INFO_PTR pInfo)
 {
-    return ock_generic_get_mechanism_info(tokdata, type, pInfo);
+    return ock_generic_get_mechanism_info(tokdata, type, pInfo, NULL);
 }
 
 CK_RV token_specific_sha_init(STDLL_TokData_t *tokdata, DIGEST_CONTEXT *ctx,
@@ -1270,6 +1294,7 @@ CK_RV token_specific_ecdh_pkcs_derive(STDLL_TokData_t *tokdata,
 CK_RV token_specific_object_add(STDLL_TokData_t * tokdata, SESSION * sess,
                                 OBJECT * obj)
 {
+    CK_ATTRIBUTE *value = NULL;
     CK_KEY_TYPE keytype;
 #ifndef NO_EC
     EVP_PKEY *ec_key = NULL;
@@ -1293,6 +1318,21 @@ CK_RV token_specific_object_add(STDLL_TokData_t * tokdata, SESSION * sess,
         return rc;
 #endif
 
+    case CKK_AES_XTS:
+        rc = template_attribute_get_non_empty(obj->template, CKA_VALUE, &value);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Failed to get CKA_VALUE\n");
+            return rc;
+        }
+
+        if (memcmp(value->pValue,
+                   ((CK_BYTE *)value->pValue) + value->ulValueLen / 2,
+                   value->ulValueLen / 2) == 0) {
+            TRACE_ERROR("The 2 key parts of an AES-XTS key can not be the same\n");
+            return CKR_ATTRIBUTE_VALUE_INVALID;
+        }
+        return CKR_OK;
+
     default:
         return CKR_OK;
     }
