@@ -97,6 +97,7 @@ struct token_specific_struct token_specific = {
     NULL,                       // generic_secret_key_gen
     // AES
     &token_specific_aes_key_gen,
+    NULL,                       // aes_xts_key_gen
     &token_specific_aes_ecb,
     &token_specific_aes_cbc,
     NULL,                       // aes_ctr
@@ -108,6 +109,7 @@ struct token_specific_struct token_specific = {
     NULL,                       // aes_cfb
     NULL,                       // aes_mac
     NULL,                       // aes_cmac
+    NULL,                       // aes_xts
     // DSA
     NULL,                       // dsa_generate_keypair
     NULL,                       // dsa_sign
@@ -121,4 +123,5 @@ struct token_specific_struct token_specific = {
     NULL,                       // set_attribute_values
     NULL,                       // set_attrs_for_new_object
     NULL,                       // handle_event
+    NULL,                       // check_obj_access
 };
