@@ -20,15 +20,17 @@
 #include <string.h>
 #include <memory.h>
 #include <dlfcn.h>
-#include <pkcs11types.h>
-#include <ep11.h>
-#include <ep11adm.h>
-#include <p11util.h>
 #include <ctype.h>
-#include <termios.h>
 #include <errno.h>
 
+#define OCK_NO_EP11_DEFINES
+#include "../../include/pkcs11types.h"
+#include "../../lib/common/p11util.h"
+#include "../../lib/ep11_stdll/ep11_func.h"
+#include "pin_prompt.h"
+
 #define EP11SHAREDLIB_NAME "OCK_EP11_LIBRARY"
+#define EP11SHAREDLIB_V4 "libep11.so.4"
 #define EP11SHAREDLIB_V3 "libep11.so.3"
 #define EP11SHAREDLIB_V2 "libep11.so.2"
 #define EP11SHAREDLIB_V1 "libep11.so.1"
@@ -41,23 +43,6 @@ CK_LONG adapter = -1;
 CK_LONG domain = -1;
 CK_OBJECT_HANDLE key_store[4096];
 
-typedef unsigned long int (*m_admin_t) (unsigned char *, size_t *,
-                                        unsigned char *,
-                                        size_t *, const unsigned char *,
-                                        size_t, const unsigned char *,
-                                        size_t, target_t);
-typedef long (*xcpa_cmdblock_t) (unsigned char *, size_t, unsigned int,
-                                 const struct XCPadmresp *,
-                                 const unsigned char *,
-                                 const unsigned char *, size_t);
-typedef long (*xcpa_internal_rv_t) (const unsigned char *, size_t,
-                                    struct XCPadmresp *, CK_RV *);
-typedef int (*m_add_module_t) (XCP_Module_t, target_t *);
-typedef int (*m_rm_module_t) (XCP_Module_t, target_t);
-typedef CK_RV (*m_get_xcp_info_t)(CK_VOID_PTR pinfo, CK_ULONG_PTR infbytes,
-                                unsigned int query, unsigned int subquery,
-                                target_t target);
-
 m_get_xcp_info_t _m_get_xcp_info;
 m_admin_t _m_admin;
 xcpa_cmdblock_t _xcpa_cmdblock;
@@ -67,8 +52,6 @@ m_rm_module_t _m_rm_module;
 
 CK_VERSION lib_version;
 
-#define CK_IBM_XCPHQ_VERSION    0xff000001
-
 typedef struct {
     short format;
     short length;
@@ -78,10 +61,7 @@ typedef struct {
 
 #define BLOBSIZE         2048*4
 
-
-
-static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
-                     CK_ULONG old_len)
+static CK_RV ep11_reencrypt(target_t target, CK_BYTE *blob, CK_ULONG blob_len)
 {
     CK_BYTE req[BLOBSIZE];
     CK_BYTE resp[BLOBSIZE];
@@ -89,6 +69,55 @@ static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
     size_t resp_len;
     struct XCPadmresp rb;
     struct XCPadmresp lrb;
+    CK_RV rc;
+
+    memset(&rb, 0, sizeof(rb));
+    memset(&lrb, 0, sizeof(lrb));
+
+    rb.domain = domain;
+    lrb.domain = domain;
+
+    resp_len = BLOBSIZE;
+
+    req_len = _xcpa_cmdblock(req, BLOBSIZE, XCP_ADM_REENCRYPT, &rb,
+                             NULL, blob, blob_len);
+
+    if (req_len < 0) {
+        fprintf(stderr, "reencrypt cmd block construction failed\n");
+        return -4;
+    }
+
+    rc = _m_admin(resp, &resp_len, NULL, 0, req, req_len, NULL, 0,
+                  target);
+    if (rc != CKR_OK || resp_len == 0) {
+        fprintf(stderr, "reencryption failed: %lx %ld\n", rc, req_len);
+        return -5;
+    }
+
+    if (_xcpa_internal_rv(resp, resp_len, &lrb, &rc) < 0) {
+        fprintf(stderr, "reencryption response malformed: %lx\n", rc);
+        return -6;
+    }
+
+    if (rc != 0) {
+        fprintf(stderr, "reencryption failed: %lx\n", rc);
+        return -7;
+    }
+
+    if (blob_len != lrb.pllen) {
+        fprintf(stderr, "reencryption blob size changed: %lx %lx %lx %lx\n",
+                blob_len, lrb.pllen, resp_len, req_len);
+        return -8;
+    }
+
+    memcpy(blob, lrb.payload, blob_len);
+
+    return 0;
+}
+
+static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj,
+                     CK_KEY_TYPE keytype, CK_BYTE *old, CK_ULONG old_len)
+{
     ep11_target_t target_list;
     struct XCP_Module module;
     target_t target = XCP_TGT_INIT;
@@ -117,8 +146,12 @@ static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
                                         1);
     }
 
-    memset(&rb, 0, sizeof(rb));
-    memset(&lrb, 0, sizeof(lrb));
+    if (rc != CKR_OK) {
+        fprintf(stderr,
+                "reencryption C_GetAttributeValue failed: obj %lx rc: %lx\n",
+                obj, rc);
+        return -1;
+    }
 
     if (_m_add_module != NULL) {
         memset(&module, 0, sizeof(module));
@@ -128,8 +161,11 @@ static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
         module.module_nr = adapter;
         XCPTGTMASK_SET_DOM(module.domainmask, domain);
         rc = _m_add_module(&module, &target);
-        if (rc != 0)
-            return CKR_FUNCTION_FAILED;
+        if (rc != 0) {
+            fprintf(stderr,
+                    "reencryption m_add_module failed: rc: %lx\n",rc);
+            return -2;
+        }
     } else {
         /* Fall back to old target handling */
         memset(&target_list, 0, sizeof(ep11_target_t));
@@ -139,53 +175,35 @@ static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
         target = (target_t)&target_list;
     }
 
-    rb.domain = domain;
-    lrb.domain = domain;
 
     fprintf(stderr, "going to reencrpyt key %lx with blob len %lx: '%s'\n", obj,
             old_len, name);
-    resp_len = BLOBSIZE;
 
-    req_len = _xcpa_cmdblock(req, BLOBSIZE, XCP_ADM_REENCRYPT, &rb,
-                              NULL, old, old_len);
-
-    if (req_len < 0) {
-        fprintf(stderr, "reencrypt cmd block construction failed\n");
-        rc = -2;
-        goto out;
-    }
-
-    rc = _m_admin(resp, &resp_len, NULL, 0, req, req_len, NULL, 0,
-                  target);
-
-    if (rc != CKR_OK || resp_len == 0) {
-        fprintf(stderr, "reencryption failed: %lx %ld\n", rc, req_len);
+    if (old_len > sizeof(blob)) {
+        fprintf(stderr,
+                "reencryption blob is too large: %lu\n", old_len);
         rc = -3;
         goto out;
     }
 
-    if (_xcpa_internal_rv(resp, resp_len, &lrb, &rc) < 0) {
-        fprintf(stderr, "reencryption response malformed: %lx\n", rc);
-        rc = -4;
-        goto out;
-    }
-
-    if (rc != 0) {
-        fprintf(stderr, "reencryption failed: %lx\n", rc);
-        rc = -7;
-        goto out;
-    }
-
-    if (old_len != lrb.pllen) {
-        fprintf(stderr, "reencryption blob size changed: %lx %lx %lx %lx\n",
-                old_len, lrb.pllen, resp_len, req_len);
-        rc = -5;
-        goto out;
-    }
-
-    memset(blob, 0, sizeof(blob));
     blob_len = old_len;
-    memcpy(blob, lrb.payload, blob_len);
+    memcpy(blob, old, blob_len);
+
+    if (keytype == CKK_AES_XTS) {
+        /* An AES XTS key consists of 2 blobs concatenated to each other */
+        rc = ep11_reencrypt(target, blob, blob_len / 2);
+        if (rc != 0)
+            goto out;
+
+        rc = ep11_reencrypt(target, blob + blob_len / 2, blob_len / 2);
+        if (rc != 0)
+            goto out;
+    } else {
+        rc = ep11_reencrypt(target, blob, blob_len);
+        if (rc != 0)
+            goto out;
+    }
+
     opaque_template[0].ulValueLen = blob_len;
 
     rc = funcs->C_SetAttributeValue(session, key_store[obj], opaque_template,
@@ -194,7 +212,7 @@ static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
         fprintf(stderr,
                 "reencryption C_SetAttributeValue failed: obj %lx '%s' rc: %lx\n",
                 obj, name, rc);
-        rc = -6;
+        rc = -9;
         goto out;
     }
 
@@ -233,7 +251,7 @@ static CK_RV get_ep11_library_version(CK_VERSION *lib_version)
 }
 
 
-static int check_card_status()
+static int check_card_status(void)
 {
     CK_RV rc;
     ep11_target_t target_list;
@@ -294,99 +312,42 @@ out:
     return rc;
 }
 
-
-int get_pin(char **pin, size_t *pinlen)
-{
-    struct termios old, new;
-    int nread;
-    char *buff = NULL;
-    size_t buflen;
-    int rc = 0;
-
-    /* turn echoing off */
-    if (tcgetattr(fileno(stdin), &old) != 0)
-        return -1;
-
-    new = old;
-    new.c_lflag &= ~ECHO;
-    if (tcsetattr(fileno(stdin), TCSAFLUSH, &new) != 0)
-        return -1;
-
-    /* read the pin
-     * Note: getline will allocate memory for buff. free it when done.
-     */
-    nread = getline(&buff, &buflen, stdin);
-    if (nread == -1) {
-        rc = -1;
-        goto done;
-    }
-
-    /* Restore terminal */
-    tcsetattr(fileno(stdin), TCSAFLUSH, &old);
-
-    /* start a newline */
-    printf("\n");
-    fflush(stdout);
-
-    /* Allocate  PIN.
-     * Note: nread includes carriage return.
-     * Replace with terminating NULL.
-     */
-    *pin = (char *) malloc(nread);
-    if (*pin == NULL) {
-        rc = -ENOMEM;
-        goto done;
-    }
-
-    /* strip the carriage return since not part of pin. */
-    buff[nread - 1] = '\0';
-    memcpy(*pin, buff, nread);
-    /* don't include the terminating null in the pinlen */
-    *pinlen = nread - 1;
-
-done:
-    if (buff)
-        free(buff);
-
-    return rc;
-}
-
 static int get_user_pin(CK_BYTE *dest)
 {
-    int ret;
-    char *userpin = NULL;
+    int ret = -1;
+    const char *userpin = NULL;
+    char *buf_user = NULL;
     size_t userpinlen;
 
-    printf("Enter the USER PIN: ");
-    fflush(stdout);
-    ret = get_pin(&userpin, &userpinlen);
-    if (ret != 0) {
+    userpin = pin_prompt(&buf_user, "Enter the USER PIN: ");
+    if (!userpin) {
         fprintf(stderr, "Could not get USER PIN.\n");
-        return -1;
+        goto out;
     }
+    userpinlen = strlen(userpin);
 
     if (userpinlen > PKCS11_MAX_PIN_LEN) {
         fprintf(stderr, "The USER PIN must be less than %d chars in length.\n",
                 (int) PKCS11_MAX_PIN_LEN);
-        free(userpin);
-        return -1;
+        goto out;
     }
 
     memcpy(dest, userpin, userpinlen + 1);
-    free(userpin);
-
-    return 0;
+    ret = 0;
+out:
+    pin_free(&buf_user);
+    return ret;
 }
 
 static int do_GetFunctionList(void)
 {
     CK_RV rc;
-    CK_RV (*func_list) () = NULL;
+    CK_RV (*func_list)(CK_FUNCTION_LIST_PTR_PTR ppFunctionList) = NULL;
     void *d;
     char *evar;
     char *evar_default = "libopencryptoki.so";
 
-    evar = getenv("PKCSLIB");
+    evar = secure_getenv("PKCSLIB");
     if (evar == NULL) {
         evar = evar_default;
     }
@@ -480,13 +441,13 @@ static int do_ParseArgs(int argc, char **argv)
 #define DLOPEN_FLAGS        RTLD_GLOBAL | RTLD_NOW
 #endif
 
-static void *ep11_load_host_lib()
+static void *ep11_load_host_lib(void)
 {
     void *lib_ep11;
     char *ep11_lib_name;
     char *errstr;
 
-    ep11_lib_name = getenv(EP11SHAREDLIB_NAME);
+    ep11_lib_name = secure_getenv(EP11SHAREDLIB_NAME);
     if (ep11_lib_name != NULL) {
         lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
 
@@ -499,9 +460,15 @@ static void *ep11_load_host_lib()
         return lib_ep11;
     }
 
-    ep11_lib_name = EP11SHAREDLIB_V3;
+    ep11_lib_name = EP11SHAREDLIB_V4;
     lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
 
+    if (lib_ep11 == NULL) {
+        /* Try version 3 instead */
+        ep11_lib_name = EP11SHAREDLIB_V3;
+        lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+    }
+
     if (lib_ep11 == NULL) {
         /* Try version 2 instead */
         ep11_lib_name = EP11SHAREDLIB_V2;
@@ -522,7 +489,7 @@ static void *ep11_load_host_lib()
 
     if (lib_ep11 == NULL) {
         errstr = dlerror();
-        fprintf(stderr, "Error loading shared library '%s[.3|.2|.1]' [%s]\n",
+        fprintf(stderr, "Error loading shared library '%s[.4|.3|.2|.1]' [%s]\n",
                 EP11SHAREDLIB, errstr);
         return NULL;
     }
@@ -684,7 +651,7 @@ int main(int argc, char **argv)
                             "rc = 0x%02x [%s]\n", rc, p11_get_ckr(rc));
                     return rc;
                 } else {
-                    if (reencrypt(session, obj,
+                    if (reencrypt(session, obj, keytype,
                                   (CK_BYTE *) opaque_template[0].pValue,
                                   opaque_template[0].ulValueLen) != 0) {
                         /* reencrypt failed */
