@@ -18,9 +18,6 @@
 #include <string.h>
 #include <memory.h>
 #include <dlfcn.h>
-#include <pkcs11types.h>
-#include <ep11.h>
-#include <p11util.h>
 #include <ctype.h>
 #include <time.h>
 #include <sys/stat.h>
@@ -29,65 +26,31 @@
 #include <regex.h>
 #include <dirent.h>
 #include <libgen.h>
-#include <termios.h>
 #include <errno.h>
 
-#define EP11SHAREDLIB_NAME "OCK_EP11_LIBRARY"
-#define EP11SHAREDLIB_V3 "libep11.so.3"
-#define EP11SHAREDLIB_V2 "libep11.so.2"
-#define EP11SHAREDLIB_V1 "libep11.so.1"
-#define EP11SHAREDLIB "libep11.so"
-#define PKCS11_MAX_PIN_LEN 128
+#define OCK_NO_EP11_DEFINES
+#include "pkcs11types.h"
+#include "defs.h"
+#include "host_defs.h"
+#include "h_extern.h"
+#include "ep11_specific.h"
+#include "pin_prompt.h"
+#include "p11util.h"
 
-#define CKH_IBM_EP11_SESSION     CKH_VENDOR_DEFINED + 1
-#define CKH_IBM_EP11_VHSMPIN     CKH_VENDOR_DEFINED + 2
-#define CKA_HIDDEN               CKA_VENDOR_DEFINED + 0x01000000
-
-#ifndef XCP_PINBLOB_BYTES
-#define  XCP_HMAC_BYTES ((size_t) (256 /8))     /* SHA-256 */
-#define  XCP_WK_BYTES   ((size_t) (256 /8))     /* keypart and session sizes  */
-#define  MOD_WRAP_BLOCKSIZE ((size_t) (128 /8)) /* blob crypt block bytecount */
-#define  XCP_PIN_SALT_BYTES  MOD_WRAP_BLOCKSIZE
-#define  XCP_PINBLOB_BYTES  \
-            (XCP_WK_BYTES +XCP_PIN_SALT_BYTES +XCP_HMAC_BYTES)
-#define  XCP_MIN_PINBYTES          8
-#define  XCP_MAX_PINBYTES         16
-#endif
-
-#define CK_IBM_XCPHQ_VERSION    0xff000001
-
-#define UNUSED(var)            ((void)(var))
-
-typedef unsigned int (*m_Logout_t) (const unsigned char *pin, size_t len,
-                                    target_t target);
-typedef int (*m_add_module_t) (XCP_Module_t module, target_t *target);
-typedef int (*m_rm_module_t) (XCP_Module_t module, target_t target);
-typedef CK_RV (*m_get_xcp_info_t)(CK_VOID_PTR pinfo, CK_ULONG_PTR infbytes,
-                                unsigned int query, unsigned int subquery,
-                                target_t target);
-
-#define SHA256_HASH_SIZE        32
 #define EP11_SESSION_ID_SIZE    16
-#define SYSFS_DEVICES_AP        "/sys/devices/ap/"
-#define REGEX_CARD_PATTERN      "card[0-9a-fA-F]+"
-#define REGEX_SUB_CARD_PATTERN  "[0-9a-fA-F]+\\.[0-9a-fA-F]+"
-#define MASK_EP11               0x04000000
 
-typedef struct {
-    short format;
-    short length;
-    short apqns[512];
-} __attribute__ ((packed)) ep11_target_t;
+#define PKCS11_MAX_PIN_LEN      128
 
 typedef CK_RV (*handler_t) (CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
                             CK_BYTE *pin_blob, CK_ULONG pin_blob_size,
                             CK_BYTE *session_id, CK_ULONG session_id_len,
                             ep11_target_t *ep11_targets,
+                            ep11_serialno_t *serial_numbers,
+                            CK_ULONG num_serial_numbers,
                             pid_t pid, CK_DATE *date);
-typedef CK_RV (*adapter_handler_t) (uint_32 adapter, uint_32 domain,
-                                    void *handler_data);
 
 CK_FUNCTION_LIST *funcs;
+m_init_t dll_m_init;
 m_Logout_t dll_m_Logout;
 m_add_module_t dll_m_add_module;
 m_rm_module_t dll_m_rm_module;
@@ -106,132 +69,87 @@ CK_VERSION lib_version;
 #define ACTION_SHOW     1
 #define ACTION_LOGOUT   2
 #define ACTION_VHSMPIN  3
-
-int get_pin(char **pin, size_t *pinlen)
-{
-    struct termios old, new;
-    int nread;
-    char *buff = NULL;
-    size_t buflen;
-    int rc = 0;
-
-    /* turn echoing off */
-    if (tcgetattr(fileno(stdin), &old) != 0)
-        return -1;
-
-    new = old;
-    new.c_lflag &= ~ECHO;
-    if (tcsetattr(fileno(stdin), TCSAFLUSH, &new) != 0)
-        return -1;
-
-    /* read the pin
-     * Note: getline will allocate memory for buff. free it when done.
-     */
-    nread = getline(&buff, &buflen, stdin);
-    if (nread == -1) {
-        rc = -1;
-        goto done;
-    }
-
-    /* Restore terminal */
-    tcsetattr(fileno(stdin), TCSAFLUSH, &old);
-
-    /* start a newline */
-    printf("\n");
-    fflush(stdout);
-
-    /* Allocate  PIN.
-     * Note: nread includes carriage return.
-     * Replace with terminating NULL.
-     */
-    *pin = (char *) malloc(nread);
-    if (*pin == NULL) {
-        rc = -ENOMEM;
-        goto done;
-    }
-
-    /* strip the carriage return since not part of pin. */
-    buff[nread - 1] = '\0';
-    memcpy(*pin, buff, nread);
-    /* don't include the terminating null in the pinlen */
-    *pinlen = nread - 1;
-
-done:
-    if (buff)
-        free(buff);
-
-    return rc;
-}
+#define ACTION_FIPSPIN  4
+#define ACTION_STATUS   5
 
 static int get_user_pin(CK_BYTE *dest)
 {
-    int ret;
-    char *userpin = NULL;
-    size_t userpinlen;
+    int ret = -1;
+    const char *userpin;
+    char *buf_user = NULL;
 
-    printf("Enter the USER PIN: ");
-    fflush(stdout);
-    ret = get_pin(&userpin, &userpinlen);
-    if (ret != 0) {
+    userpin = pin_prompt(&buf_user, "Enter the USER PIN: ");
+    if (!userpin) {
         fprintf(stderr, "Could not get USER PIN.\n");
-        return -1;
+        goto out;
     }
 
-    if (userpinlen > PKCS11_MAX_PIN_LEN) {
+    if (strlen(userpin) > PKCS11_MAX_PIN_LEN) {
         fprintf(stderr, "The USER PIN must be less than %d chars in length.\n",
                 (int) PKCS11_MAX_PIN_LEN);
-        free(userpin);
-        return -1;
+        goto out;
     }
 
-    memcpy(dest, userpin, userpinlen + 1);
-    free(userpin);
-
-    return 0;
+    memcpy(dest, userpin, strlen(userpin) + 1);
+    ret = 0;
+out:
+    pin_free(&buf_user);
+    return ret;
 }
 
-static int get_vhsm_pin(CK_BYTE *dest)
+static int get_pin(CK_BYTE *dest, const char *name)
 {
-    int ret;
-    char *vhsmpin = NULL;
-    size_t vhsmpinlen;
+    int ret = -1;
+    const char *pin = NULL;
+    char *buf = NULL;
+    size_t pinlen;
+    char *msg1 = NULL, *msg2 = NULL;
 
-    printf("Enter the new VHSM PIN: ");
-    fflush(stdout);
-    ret = get_pin(&vhsmpin, &vhsmpinlen);
-    if (ret != 0) {
-        fprintf(stderr, "Could not get VHSM PIN.\n");
-        return -1;
+    if (asprintf(&msg1, "Enter the new %s: ", name) < 0 ||
+        asprintf(&msg2, "Re-enter the new %s: ", name) < 0) {
+        fprintf(stderr, "Failed to allocate memory.\n");
+        goto out;
     }
 
-    if (vhsmpinlen < XCP_MIN_PINBYTES) {
-        fprintf(stderr, "The VHSM PIN must be at least %d chars in length.\n",
+    pin = pin_prompt_new(&buf, msg1, msg2);
+    if (!pin) {
+        fprintf(stderr, "Could not get %sN.\n", name);
+        goto out;
+    }
+    pinlen = strlen(pin);
+
+    if (pinlen < XCP_MIN_PINBYTES) {
+        fprintf(stderr, "The %s must be at least %d chars in length.\n", name,
                 (int) XCP_MIN_PINBYTES);
-        free(vhsmpin);
-        return -1;
+        goto out;
     }
-    if (vhsmpinlen > XCP_MAX_PINBYTES) {
-        fprintf(stderr, "The VHSM PIN must be less than %d chars in length.\n",
+    if (pinlen > XCP_MAX_PINBYTES) {
+        fprintf(stderr, "The %s must be less than %d chars in length.\n", name,
                 (int) XCP_MAX_PINBYTES);
-        free(vhsmpin);
-        return -1;
+        goto out;
     }
 
-    memcpy(dest, vhsmpin, vhsmpinlen + 1);
-    free(vhsmpin);
+    memcpy(dest, pin, pinlen + 1);
+    ret = 0;
+out:
+    pin_free(&buf);
+    if (msg1 != NULL)
+        free(msg1);
+    if(msg2 != NULL)
+        free(msg2);
 
-    return 0;
+    return ret;
 }
 
 static int do_GetFunctionList(void)
 {
     CK_RV rc;
-    CK_RV (*func_list)() = NULL;
+    CK_RV (*func_list)(CK_FUNCTION_LIST_PTR_PTR ppFunctionList) = NULL;
     void *d;
     char *evar;
     char *evar_default = "libopencryptoki.so";
 
-    evar = getenv("PKCSLIB");
+    evar = secure_getenv("PKCSLIB");
     if (evar == NULL)
         evar = evar_default;
 
@@ -265,8 +183,8 @@ int is_ep11_token(CK_SLOT_ID slot_id)
 
 static void usage(char *fct)
 {
-    printf("usage:  %s show|logout|vhsmpin [-date <yyyy/mm/dd>] [-pid <pid>] "
-           "[-id <sess-id>] [-slot <num>] [-force] [-h]\n\n", fct);
+    printf("usage:  %s show|logout|vhsmpin|fipspin|status [-date <yyyy/mm/dd>] "
+           "[-pid <pid>] [-id <sess-id>] [-slot <num>] [-force] [-h]\n\n", fct);
     return;
 }
 
@@ -292,6 +210,10 @@ static int do_ParseArgs(int argc, char **argv)
         action = ACTION_LOGOUT;
     } else if (strcmp(argv[1], "vhsmpin") == 0) {
         action = ACTION_VHSMPIN;
+    } else if (strcmp(argv[1], "fipspin") == 0) {
+        action = ACTION_FIPSPIN;
+    } else if (strcmp(argv[1], "status") == 0) {
+        action = ACTION_STATUS;
     } else {
         printf("Unknown Action given. For help use the '--help' or '-h' "
                "option.\n");
@@ -302,6 +224,10 @@ static int do_ParseArgs(int argc, char **argv)
         if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {
             usage(argv[0]);
             return 0;
+        } else if (action == ACTION_STATUS) {
+            printf("Argument '%s' not accepted for 'status' command\n",
+                   argv[i]);
+            return -1;
         } else if (strcmp(argv[i], "-slot") == 0) {
             if (argc <= i + 1 || !isdigit(*argv[i + 1])) {
                 printf("Slot parameter is not numeric!\n");
@@ -357,7 +283,7 @@ static int do_ParseArgs(int argc, char **argv)
             return -1;
         }
     }
-    if (SLOT_ID == (CK_SLOT_ID)(-1)) {
+    if (action != ACTION_STATUS && SLOT_ID == (CK_SLOT_ID)(-1)) {
         printf("Slot-ID not set!\n");
         return -1;
     }
@@ -405,11 +331,65 @@ static CK_RV get_ep11_library_version(CK_VERSION *lib_version)
     return CKR_OK;
 }
 
+static CK_RV get_serial_number(target_t target, ep11_serialno_t serial_number)
+{
+    CK_IBM_XCP_INFO xcp_info;
+    CK_ULONG xcp_info_len = sizeof(xcp_info);
+    CK_RV rc;
+
+    rc = dll_m_get_xcp_info(&xcp_info, &xcp_info_len, CK_IBM_XCPQ_MODULE, 0,
+                            target);
+    if (rc != CKR_OK)
+        return rc;
+
+    memcpy(serial_number, xcp_info.serialNumber, sizeof(ep11_serialno_t));
+
+    return CKR_OK;
+}
+
+static void remove_serial_number(ep11_serialno_t serial_number,
+                                 ep11_serialno_t *serial_numbers,
+                                 CK_ULONG *num_serial_numbers)
+{
+    CK_ULONG i;
+    CK_BOOL found = FALSE;
+
+    for(i = 0; i < *num_serial_numbers; i++) {
+        if (memcmp(serial_numbers[i], serial_number,
+                   sizeof(ep11_serialno_t)) == 0) {
+            found = TRUE;
+            break;
+        }
+    }
+
+    if (!found)
+        return;
+
+    memmove(serial_numbers[i],
+            serial_numbers[i + 1],
+            (*num_serial_numbers - i) * sizeof(ep11_serialno_t));
+    memset(serial_numbers[*num_serial_numbers - 1],
+           0, sizeof(ep11_serialno_t));
+
+   (*num_serial_numbers)--;
+
+    return;
+}
+
+struct logout_data {
+    CK_BYTE *pin_blob;
+    CK_ULONG pin_blob_size;
+    ep11_serialno_t *serial_numbers;
+    CK_ULONG num_serial_numbers;
+};
+
 static CK_RV logout_handler(uint_32 adapter, uint_32 domain, void *handler_data)
 {
     ep11_target_t target_list;
     struct XCP_Module module;
     target_t target = XCP_TGT_INIT;
+    struct logout_data *data = handler_data;
+    ep11_serialno_t serial_number;
     CK_RV rc;
 
     if (dll_m_add_module != NULL) {
@@ -431,14 +411,28 @@ static CK_RV logout_handler(uint_32 adapter, uint_32 domain, void *handler_data)
         target = (target_t)&target_list;
     }
 
-    rc = dll_m_Logout(handler_data, XCP_PINBLOB_BYTES, target);
+    rc = get_serial_number(target, serial_number);
+    if (rc != CKR_OK) {
+        fprintf(stderr,
+                "WARNING: Querying adapter %02X.%04X failed: 0x%lx [%s]\n",
+                adapter, domain, rc, p11_get_ckr(rc));
+        error = rc;
+        goto out;
+    }
+
+    rc = dll_m_Logout(data->pin_blob, data->pin_blob_size, target);
     if (rc != CKR_OK && rc != CKR_SESSION_CLOSED) {
         fprintf(stderr,
                 "WARNING: Logout failed for adapter %02X.%04X: 0x%lx [%s]\n",
                 adapter, domain, rc, p11_get_ckr(rc));
         error = rc;
+        goto out;
     }
 
+    remove_serial_number(serial_number,
+                         data->serial_numbers, &data->num_serial_numbers);
+
+out:
     if (dll_m_rm_module != NULL)
         dll_m_rm_module(&module, target);
 
@@ -499,6 +493,16 @@ static CK_RV is_card_ep11_and_online(const char *name)
     if (strcmp(buf, "1") != 0)
         return CKR_FUNCTION_FAILED;
 
+    sprintf(fname, "%s%s/config", SYSFS_DEVICES_AP, name);
+    rc = file_fgets(fname, buf, sizeof(buf));
+    if (rc == CKR_OK && strcmp(buf, "1") != 0)
+        return CKR_FUNCTION_FAILED;
+
+    sprintf(fname, "%s%s/chkstop", SYSFS_DEVICES_AP, name);
+    rc = file_fgets(fname, buf, sizeof(buf));
+    if (rc == CKR_OK && strcmp(buf, "0") != 0)
+        return CKR_FUNCTION_FAILED;
+
     sprintf(fname, "%s%s/ap_functions", SYSFS_DEVICES_AP, name);
     rc = file_fgets(fname, buf, sizeof(buf));
     if (rc != CKR_OK)
@@ -615,9 +619,8 @@ static CK_RV scan_for_ep11_cards(adapter_handler_t handler, void *handler_data)
     return CKR_OK;
 }
 
-static CK_RV handle_all_ep11_cards(ep11_target_t *ep11_targets,
-                                   adapter_handler_t handler,
-                                   void *handler_data)
+CK_RV handle_all_ep11_cards(ep11_target_t * ep11_targets,
+                            adapter_handler_t handler, void *handler_data)
 {
     int i;
     CK_RV rc;
@@ -638,17 +641,14 @@ static CK_RV handle_all_ep11_cards(ep11_target_t *ep11_targets,
     return CKR_OK;
 }
 
-static CK_RV logout_session_obj(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
-                                CK_BYTE *pin_blob, CK_ULONG pin_blob_size,
-                                CK_BYTE *session_id, CK_ULONG session_id_len,
+static void print_session_info(CK_BYTE *session_id, CK_ULONG session_id_len,
                                 ep11_target_t *ep11_targets,
+                                ep11_serialno_t *serial_numbers,
+                                CK_ULONG num_serial_numbers,
                                 pid_t pid, CK_DATE *date)
 {
-    CK_RV rc;
     CK_ULONG i;
 
-    UNUSED(pin_blob_size);
-
     for (i = 0; i < session_id_len; i++)
         printf("%02X", session_id[i]);
     printf(":\n");
@@ -657,6 +657,75 @@ static CK_RV logout_session_obj(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
     else
         printf("\tPid:\t%u\n", pid);
     printf("\tDate:\t%.4s/%.2s/%.2s\n", date->year, date->month, date->day);
+    if (ep11_targets->length == 0) {
+        printf("\tAPQNs:\tANY\n");
+    } else {
+        printf("\tAPQNs:\n");
+        for (i = 0; i < (CK_ULONG)ep11_targets->length; i++)
+            printf("\t\t%02X.%04X\n", ep11_targets->apqns[2 * i],
+                   ep11_targets->apqns[2 * i + 1]);
+    }
+    if (num_serial_numbers) {
+        printf("\tAdapter serial numbers:\n");
+        for (i = 0; i < num_serial_numbers; i++)
+            printf("\t\t%.16s\n", serial_numbers[i]);
+    }
+}
+
+static CK_RV update_ep11_object(CK_SESSION_HANDLE session,
+                                CK_OBJECT_HANDLE obj,
+                                ep11_target_t *ep11_targets,
+                                ep11_serialno_t *serial_numbers,
+                                CK_ULONG num_serial_numbers)
+{;
+    CK_RV rc;
+    CK_BYTE *app_data = NULL;
+    CK_ULONG app_data_len;
+    CK_ATTRIBUTE attr = { CKA_APPLICATION, NULL, 0 };
+
+    app_data_len = sizeof(ep11_target_t) +
+                                num_serial_numbers * sizeof(ep11_serialno_t);
+    app_data = malloc(app_data_len);
+    if (app_data == NULL) {
+        fprintf(stderr, "Failed to allocate memory\n");
+        rc = CKR_HOST_MEMORY;
+        goto out;
+    }
+
+    memcpy(app_data, ep11_targets, sizeof(ep11_target_t));
+    if (num_serial_numbers > 0)
+        memcpy(app_data + sizeof(ep11_target_t), serial_numbers,
+                                num_serial_numbers * sizeof(ep11_serialno_t));
+    attr.pValue = app_data;
+    attr.ulValueLen = app_data_len;
+
+    rc = funcs->C_SetAttributeValue(session, obj, &attr, 1);
+    if (rc != CKR_OK) {
+        fprintf(stderr, "C_SetAttributeValue() rc = 0x%02lx [%s]\n", rc,
+               p11_get_ckr(rc));
+    }
+
+out:
+    if (app_data != NULL)
+        free(app_data);
+
+    return rc;
+}
+
+static CK_RV logout_session_obj(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
+                                CK_BYTE *pin_blob, CK_ULONG pin_blob_size,
+                                CK_BYTE *session_id, CK_ULONG session_id_len,
+                                ep11_target_t *ep11_targets,
+                                ep11_serialno_t *serial_numbers,
+                                CK_ULONG num_serial_numbers,
+                                pid_t pid, CK_DATE *date)
+{
+    CK_RV rc;
+    struct logout_data data;
+    CK_ULONG i;
+
+    print_session_info(session_id, session_id_len, ep11_targets,
+                       serial_numbers, num_serial_numbers, pid, date);
 
     if (is_process_running(pid)) {
         printf("\tSession is not logged out, process %u is still running\n",
@@ -665,36 +734,58 @@ static CK_RV logout_session_obj(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
     }
 
     error = CKR_OK;
-    rc = handle_all_ep11_cards(ep11_targets, logout_handler, pin_blob);
+    data.pin_blob = pin_blob;
+    data.pin_blob_size = pin_blob_size;
+    data.serial_numbers = serial_numbers;
+    data.num_serial_numbers = num_serial_numbers;
+    rc = handle_all_ep11_cards(ep11_targets, logout_handler, &data);
     if (rc != CKR_OK) {
         fprintf(stderr, "handle_all_ep11_cards() rc = 0x%02lx [%s]\n", rc,
                 p11_get_ckr(rc));
         return rc;
     }
+
     if (error != CKR_OK) {
         fprintf(stderr,
-                "WARNING: Not all APQNs were successfully logged out.\n");
-        if (!force) {
+                "\tWARNING: Not all APQNs were successfully logged out.\n");
+        if (!force)
             fprintf(stderr,
-                    "         Session is not deleted. Specify -force to delete"
-                    "it anyway.\n");
+                    "\t         Session is not deleted. Specify -force to "
+                    "delete it anyway.\n");
+    }
+
+    if ((error == CKR_OK && data.num_serial_numbers == 0) || force) {
+        rc = funcs->C_DestroyObject(session, obj);
+        if (rc != CKR_OK) {
+            fprintf(stderr, "C_DestroyObject() rc = 0x%02lx [%s]\n", rc,
+                    p11_get_ckr(rc));
             return rc;
         }
+    } else {
+        rc = update_ep11_object(session, obj, ep11_targets,
+                                data.serial_numbers, data.num_serial_numbers);
+        if (rc != CKR_OK)
+            return rc;
     }
 
-    rc = funcs->C_DestroyObject(session, obj);
-    if (rc != CKR_OK) {
-        fprintf(stderr, "C_DestroyObject() rc = 0x%02lx [%s]\n", rc,
-                p11_get_ckr(rc));
-        return rc;
-    }
-
-    if (!force)
-        printf("\tSession logged out successfully\n");
-    else
+    if (!force) {
+        if (data.num_serial_numbers == 0) {
+            printf("\tSession logged out successfully\n");
+            count++;
+        } else {
+            fprintf(stderr,
+                    "\tWARNING: Session not logged out on the adapters with\n"
+                    "\t         the following serial numbers:\n");
+            for (i = 0; i < data.num_serial_numbers; i++)
+                printf("\t\t\t%.16s\n", data.serial_numbers[i]);
+            fprintf(stderr,
+                    "\t         Session is not deleted. Specify -force to "
+                    "delete it anyway.\n");
+        }
+    } else {
         printf("\tSession deleted due to -force option\n");
-
-    count++;
+        count++;
+    }
 
     return CKR_OK;
 }
@@ -705,24 +796,17 @@ static CK_RV show_session_obj(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
                               CK_BYTE *pin_blob, CK_ULONG pin_blob_size,
                               CK_BYTE *session_id, CK_ULONG session_id_len,
                               ep11_target_t *ep11_targets,
+                              ep11_serialno_t *serial_numbers,
+                              CK_ULONG num_serial_numbers,
                               pid_t pid, CK_DATE *date)
 {
-    CK_ULONG i;
-
     UNUSED(session);
     UNUSED(obj);
     UNUSED(pin_blob);
     UNUSED(pin_blob_size);
-    UNUSED(ep11_targets);
 
-    for (i = 0; i < session_id_len; i++)
-        printf("%02X", session_id[i]);
-    printf(":\n");
-    if (is_process_running(pid))
-        printf("\tPid:\t%u (still running)\n", pid);
-    else
-        printf("\tPid:\t%u\n", pid);
-    printf("\tDate:\t%.4s/%.2s/%.2s\n", date->year, date->month, date->day);
+    print_session_info(session_id, session_id_len, ep11_targets,
+                       serial_numbers, num_serial_numbers, pid, date);
 
     count++;
 
@@ -780,13 +864,17 @@ static CK_RV process_session_obj(CK_SESSION_HANDLE session,
     CK_BBOOL match;
     CK_BYTE pin_blob[XCP_PINBLOB_BYTES];
     CK_BYTE session_id[EP11_SESSION_ID_SIZE];
-    ep11_target_t ep11_targets;
+    struct {
+        ep11_target_t ep11_targets;
+        ep11_serialno_t serial_numbers[MAX_APQN];
+    } app_data;
+    CK_ULONG num_serial_numbers = 0;
     pid_t pid;
     CK_DATE date;
     CK_ATTRIBUTE attrs[] = {
         { CKA_VALUE, pin_blob, sizeof(pin_blob) },
         { CKA_ID, session_id, sizeof(session_id) },
-        { CKA_APPLICATION, &ep11_targets, sizeof(ep11_targets) },
+        { CKA_APPLICATION, &app_data, sizeof(app_data) }, // keep at index 2
         { CKA_OWNER, &pid, sizeof(pid) },
         { CKA_START_DATE, &date, sizeof(date) },
     };
@@ -802,6 +890,9 @@ static CK_RV process_session_obj(CK_SESSION_HANDLE session,
         return CKR_OK;
     }
 
+    num_serial_numbers = (attrs[2].ulValueLen - sizeof(ep11_target_t)) /
+                                                    sizeof(ep11_serialno_t);
+
     /* Ignore our own EP11 session */
     if (pid == getpid())
         return CKR_OK;
@@ -810,7 +901,9 @@ static CK_RV process_session_obj(CK_SESSION_HANDLE session,
 
     if (match) {
         rc = handler(session, obj, pin_blob, sizeof(pin_blob),
-                     session_id, sizeof(session_id), &ep11_targets, pid, &date);
+                     session_id, sizeof(session_id), &app_data.ep11_targets,
+                     app_data.serial_numbers, num_serial_numbers,
+                     pid, &date);
         if (rc != CKR_OK)
             return rc;
     }
@@ -826,12 +919,12 @@ static CK_RV find_sessions(CK_SESSION_HANDLE session, handler_t handler)
     CK_ULONG obj;
     CK_OBJECT_CLASS class = CKO_HW_FEATURE;
     CK_HW_FEATURE_TYPE type = CKH_IBM_EP11_SESSION;
-    CK_BYTE true = TRUE;
+    CK_BYTE ck_true = TRUE;
     CK_ATTRIBUTE session_template[] = {
         { CKA_CLASS, &class, sizeof(class) },
-        { CKA_TOKEN, &true, sizeof(true) },
-        { CKA_PRIVATE, &true, sizeof(true) },
-        { CKA_HIDDEN, &true, sizeof(true) },
+        { CKA_TOKEN, &ck_true, sizeof(ck_true) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
         { CKA_HW_FEATURE_TYPE, &type, sizeof(type) },
     };
 
@@ -892,20 +985,19 @@ static CK_RV logout_sessions(CK_SESSION_HANDLE session)
     return rc;
 }
 
-static CK_RV find_vhsmpin_object(CK_SESSION_HANDLE session,
-                                 CK_OBJECT_HANDLE *obj)
+static CK_RV find_pin_object(CK_SESSION_HANDLE session, CK_HW_FEATURE_TYPE type,
+                             CK_OBJECT_HANDLE *obj)
 {
     CK_RV rc;
     CK_OBJECT_HANDLE obj_store[16];
     CK_ULONG objs_found = 0;
     CK_OBJECT_CLASS class = CKO_HW_FEATURE;
-    CK_HW_FEATURE_TYPE type = CKH_IBM_EP11_VHSMPIN;
-    CK_BYTE true = TRUE;
+    CK_BBOOL ck_true = TRUE;
     CK_ATTRIBUTE vhsmpin_template[] = {
         { CKA_CLASS, &class, sizeof(class) },
-        { CKA_TOKEN, &true, sizeof(true) },
-        { CKA_PRIVATE, &true, sizeof(true) },
-        { CKA_HIDDEN, &true, sizeof(true) },
+        { CKA_TOKEN, &ck_true, sizeof(ck_true) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
         { CKA_HW_FEATURE_TYPE, &type, sizeof(type) },
     };
 
@@ -937,35 +1029,46 @@ out:
     return rc;
 }
 
+static const char *hwtype2str(CK_HW_FEATURE_TYPE type) {
+    switch (type) {
+    case CKH_IBM_EP11_VHSMPIN:
+        return "VHSM PIN";
+    case CKH_IBM_EP11_FIPSPIN:
+        return "FIPS PIN";
+    default:
+        return "unknown";
+    }
+}
 
-static CK_RV set_vhsmpin(CK_SESSION_HANDLE session)
+static CK_RV set_pin(CK_SESSION_HANDLE session, CK_HW_FEATURE_TYPE type)
 {
     CK_RV rc;
-    CK_BYTE vhsm_pin[XCP_MAX_PINBYTES + 1];
+    CK_BYTE pin[XCP_MAX_PINBYTES + 1];
     CK_OBJECT_HANDLE obj = CK_INVALID_HANDLE;
     CK_OBJECT_CLASS class = CKO_HW_FEATURE;
-    CK_HW_FEATURE_TYPE type = CKH_IBM_EP11_VHSMPIN;
-    CK_BYTE subject[] = "EP11 VHSM-Pin Object";
-    CK_BYTE true = TRUE;
+    char subject[200];
+    CK_BBOOL ck_true = TRUE;
 
-    if (get_vhsm_pin(vhsm_pin)) {
-        fprintf(stderr, "get_vhsm_pin() failed\n");
+    if (get_pin(pin, hwtype2str(type))) {
+        fprintf(stderr, "get_pin() failed\n");
         return CKR_FUNCTION_FAILED;
     }
 
+    snprintf(subject, sizeof(subject), "EP11 %s Object", hwtype2str(type));
+
     CK_ATTRIBUTE attrs[] = {
         { CKA_CLASS, &class, sizeof(class) },
-        { CKA_TOKEN, &true, sizeof(true) },
-        { CKA_PRIVATE, &true, sizeof(true) },
-        { CKA_HIDDEN, &true, sizeof(true) },
+        { CKA_TOKEN, &ck_true, sizeof(ck_true) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
         { CKA_HW_FEATURE_TYPE, &type, sizeof(type) },
-        { CKA_SUBJECT, &subject, sizeof(subject) },
-        { CKA_VALUE, vhsm_pin, strlen((char *)vhsm_pin) },
+        { CKA_SUBJECT, &subject, strlen(subject) },
+        { CKA_VALUE, pin, strlen((char *)pin) },
     };
 
-    rc = find_vhsmpin_object(session, &obj);
+    rc = find_pin_object(session, type, &obj);
     if (rc != CKR_OK) {
-        fprintf(stderr, "find_vhsmpin_object() failed\n");
+        fprintf(stderr, "find_pin_object() failed\n");
         return CKR_FUNCTION_FAILED;
     }
 
@@ -986,7 +1089,109 @@ static CK_RV set_vhsmpin(CK_SESSION_HANDLE session)
                 p11_get_ckr(rc));
         return rc;
     }
-    printf("VHSM-pin successfully set.\n");
+    printf("%s successfully set.\n", hwtype2str(type));
+
+    return CKR_OK;
+}
+
+static CK_RV status_handler(uint_32 adapter, uint_32 domain,
+                            void *handler_data)
+{
+    ep11_target_t target_list;
+    struct XCP_Module module;
+    target_t target = XCP_TGT_INIT;
+    uint32_t *res = NULL;
+    CK_ULONG reslen = 0;
+    CK_ULONG i;
+    uint32_t caps = 0;
+    CK_RV rc;
+    CK_BBOOL found = CK_FALSE;
+
+    UNUSED(handler_data);
+
+    if (dll_m_add_module != NULL) {
+        memset(&module, 0, sizeof(module));
+        module.version = lib_version.major >= 3 ? XCP_MOD_VERSION_2
+                                                : XCP_MOD_VERSION_1;
+        module.flags = XCP_MFL_MODULE;
+        module.module_nr = adapter;
+        XCPTGTMASK_SET_DOM(module.domainmask, domain);
+        rc = dll_m_add_module(&module, &target);
+        if (rc != 0) {
+            fprintf(stderr,
+                    "dll_m_add_module (EXT_CAPLIST) failed: rc=0x%lx\n", rc);
+            return CKR_FUNCTION_FAILED;
+        }
+    } else {
+        /* Fall back to old target handling */
+        memset(&target_list, 0, sizeof(ep11_target_t));
+        target_list.length = 1;
+        target_list.apqns[0] = adapter;
+        target_list.apqns[1] = domain;
+        target = (target_t)&target_list;
+    }
+
+    printf("APQN %02x.%04x:\n", adapter, domain);
+
+    reslen = sizeof(caps);
+    rc = dll_m_get_xcp_info(&caps, &reslen, CK_IBM_XCPQ_EXT_CAPS, 0, target);
+    if (rc != CKR_OK || reslen != sizeof(caps)) {
+        fprintf(stderr,
+                "dll_m_get_xcp_info (EXT_CAPS) failed: rc=0x%lx\n", rc);
+        goto done;
+    }
+
+    if (caps == 0)
+        goto no_info;
+
+    reslen = caps * sizeof(uint32_t) * 2;
+    res = calloc(1, reslen);
+    if (res == NULL) {
+        fprintf(stderr, "Failed to allocate memory\n");
+        goto done;
+    }
+
+    rc = dll_m_get_xcp_info(res, &reslen, CK_IBM_XCPQ_EXT_CAPLIST, 0, target);
+    if (rc != CKR_OK) {
+        fprintf(stderr,
+                "dll_m_get_xcp_info (EXT_CAPLIST) failed: rc=0x%lx\n", rc);
+        goto done;
+    }
+
+    for (i = 0; i < reslen / 4; i += 2) {
+        if (res[i] == CK_IBM_XCPXQ_MAX_SESSIONS) {
+            printf("  Max Sessions:        %u\n", res[i + 1]);
+            found = CK_TRUE;
+        } else if (res[i] == CK_IBM_XCPXQ_AVAIL_SESSIONS) {
+            printf("  Available Sessions:  %u\n", res[i + 1]);
+            found = CK_TRUE;
+        }
+    }
+
+no_info:
+    if (!found)
+        printf("  Information not available\n");
+
+done:
+    if (dll_m_rm_module != NULL)
+        dll_m_rm_module(&module, target);
+    if (res != NULL)
+        free(res);
+
+    return CKR_OK;
+}
+
+static CK_RV show_ep11_status(void)
+{
+    ep11_target_t any_target = { 0, 0, { 0 } };
+    CK_RV rc;
+
+    rc = handle_all_ep11_cards(&any_target, status_handler, NULL);
+    if (rc != CKR_OK) {
+        fprintf(stderr, "handle_all_ep11_cards() rc = 0x%02lx [%s]\n", rc,
+                p11_get_ckr(rc));
+        return rc;
+    }
 
     return CKR_OK;
 }
@@ -997,13 +1202,13 @@ static CK_RV set_vhsmpin(CK_SESSION_HANDLE session)
 #define DLOPEN_FLAGS        RTLD_GLOBAL | RTLD_NOW
 #endif
 
-static void *ep11_load_host_lib()
+static void *ep11_load_host_lib(void)
 {
     void *lib_ep11;
     char *ep11_lib_name;
     char *errstr;
 
-    ep11_lib_name = getenv(EP11SHAREDLIB_NAME);
+    ep11_lib_name = secure_getenv(EP11SHAREDLIB_NAME);
     if (ep11_lib_name != NULL) {
         lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
 
@@ -1016,9 +1221,15 @@ static void *ep11_load_host_lib()
         return lib_ep11;
     }
 
-    ep11_lib_name = EP11SHAREDLIB_V3;
+    ep11_lib_name = EP11SHAREDLIB_V4;
     lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
 
+    if (lib_ep11 == NULL) {
+        /* Try version 3 instead */
+        ep11_lib_name = EP11SHAREDLIB_V3;
+        lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+    }
+
     if (lib_ep11 == NULL) {
         /* Try version 2 instead */
         ep11_lib_name = EP11SHAREDLIB_V2;
@@ -1039,7 +1250,7 @@ static void *ep11_load_host_lib()
 
     if (lib_ep11 == NULL) {
         errstr = dlerror();
-        fprintf(stderr, "Error loading shared library '%s[.3|.2|.1]' [%s]\n",
+        fprintf(stderr, "Error loading shared library '%s[.4|.3|.2|.1]' [%s]\n",
                 EP11SHAREDLIB, errstr);
         return NULL;
     }
@@ -1066,9 +1277,11 @@ int main(int argc, char **argv)
     if (!lib_ep11)
         return CKR_FUNCTION_FAILED;
 
+    *(void **)(&dll_m_init) = dlsym(lib_ep11, "m_init");
     *(void **)(&dll_m_Logout) = dlsym(lib_ep11, "m_Logout");
     *(void **)(&dll_m_get_xcp_info) = dlsym(lib_ep11, "m_get_xcp_info");
-    if (dll_m_Logout == NULL || dll_m_get_xcp_info == NULL) {
+    if (dll_m_init == NULL || dll_m_Logout == NULL ||
+        dll_m_get_xcp_info == NULL) {
         fprintf(stderr, "ERROR loading shared lib '%s' [%s]\n",
                 EP11SHAREDLIB, dlerror());
         return CKR_FUNCTION_FAILED;
@@ -1085,10 +1298,21 @@ int main(int argc, char **argv)
         dll_m_rm_module = NULL;
     }
 
+    rc = dll_m_init();
+    if (rc != CKR_OK) {
+        fprintf(stderr, "ERROR dll_m_init() Failed, rx = 0x%0x\n", rc);
+        return rc;
+    }
+
     rc = get_ep11_library_version(&lib_version);
     if (rc != CKR_OK)
         return rc;
 
+    if (action == ACTION_STATUS) {
+        rc = show_ep11_status();
+        return rc;
+    }
+
     printf("Using slot #%lu...\n\n", SLOT_ID);
 
     rc = do_GetFunctionList();
@@ -1105,16 +1329,17 @@ int main(int argc, char **argv)
         CK_SESSION_HANDLE hsess = 0;
         rc = funcs->C_GetFunctionStatus(hsess);
         if (rc != CKR_FUNCTION_NOT_PARALLEL)
-            return rc;
+            goto done;
 
         rc = funcs->C_CancelFunction(hsess);
         if (rc != CKR_FUNCTION_NOT_PARALLEL)
-            return rc;
+            goto done;
     }
 
     if (!is_ep11_token(SLOT_ID)) {
         fprintf(stderr, "ERROR Slot %lu is not an EP11 token\n", SLOT_ID);
-        return CKR_FUNCTION_FAILED;
+        rc = CKR_FUNCTION_FAILED;
+        goto done;
     }
 
     flags = CKF_SERIAL_SESSION | CKF_RW_SESSION;
@@ -1123,7 +1348,7 @@ int main(int argc, char **argv)
         fprintf(stderr, "C_OpenSession() rc = 0x%02x [%s]\n", rc,
                 p11_get_ckr(rc));
         session = CK_INVALID_HANDLE;
-        return rc;
+        goto done;
     }
 
     if (get_user_pin(user_pin)) {
@@ -1132,14 +1357,14 @@ int main(int argc, char **argv)
         if (rc != CKR_OK)
             fprintf(stderr, "C_CloseAllSessions() rc = 0x%02x [%s]\n", rc,
                     p11_get_ckr(rc));
-        return rc;
+        goto done;
     }
 
     user_pin_len = (CK_ULONG) strlen((char *) user_pin);
     rc = funcs->C_Login(session, CKU_USER, user_pin, user_pin_len);
     if (rc != CKR_OK) {
         fprintf(stderr, "C_Login() rc = 0x%02x [%s]\n", rc, p11_get_ckr(rc));
-        return rc;
+        goto done;
     }
 
     switch (action) {
@@ -1150,14 +1375,17 @@ int main(int argc, char **argv)
         rc = logout_sessions(session);
         break;
     case ACTION_VHSMPIN:
-        rc = set_vhsmpin(session);
+        rc = set_pin(session, CKH_IBM_EP11_VHSMPIN);
+        break;
+    case ACTION_FIPSPIN:
+        rc = set_pin(session, CKH_IBM_EP11_FIPSPIN);
         break;
     }
-    if (rc != CKR_OK)
-        return rc;
 
-    rc = funcs->C_Logout(session);
-    rc = funcs->C_CloseAllSessions(SLOT_ID);
+    funcs->C_Logout(session);
+done:
+    funcs->C_CloseAllSessions(SLOT_ID);
+    funcs->C_Finalize(NULL);
 
     return rc;
 }
